from .Model import Model, MC_ID, MDP_ID, CTMC_ID
from numpy import ndarray, where
from random import choices

class Base_MC(Model):
	"""
	Abstract class that represents a model.
	Is inherited by MC, CTMC and MDP.
	Should not be instantiated itself!
	"""
	def __init__(self, matrix: ndarray, labelling: list, name: str) -> None:
		"""
		Creates an abstract model for MC, CTMC and MDP.

		Parameters
		----------
		matrix : ndarray
			Represents the transition matrix.
		labelling: list of str
			A list of N observations (with N the nb of states).
			If `labelling[s] == o` then state of ID `s` is labelled by `o`.
			Each state has exactly one label.
		name : str, optional
			Name of the model.
			Default is "unknow_MC"
		"""
		self.labelling = labelling
		self.alphabet = list(set(labelling))
		
		if not 'init' in self.labelling:
			msg = "No initial state given: at least one"
			msg += " state should be labelled by 'init'."
			raise ValueError(msg)
		initial_state = [1.0/self.labelling.count("init") if i=='init' else 0.0 for i in self.labelling]

		super().__init__(matrix,initial_state,name)
		if len(self.labelling) != self.nb_states:
			raise ValueError("The length of labelling ("+str(len(labelling))+") is not equal to the number of states("+str(self.nb_states)+")")
	
	def getLabel(self,state: int) -> str:
		"""
		Returns the label of `state`.

		Parameters
		----------
		state : int
			a state ID

		Returns
		-------
		str
			a label

		Example
		-------
		>>> model.getLabel(2)
		'Label-of-state-2'
		"""
		self._checkStateIndex(state)
		return self.labelling[state]

	def b(self,state:int, label:str):
		if self.getLabel(state) == label:
			return 1.0
		return 0.0
	
	def getAlphabet(self) -> list:
		"""
		Returns the alphabet of this model.

		Returns
		-------
		list of str
			The alphabet of this model
		
		Example
		-------
		>>> model.getAlphabet()
		['a','b','c','d','done']
		"""
		return self.alphabet

	def _savePrism(self,f) -> None:
		f.write("module "+self.name+'\n')
		f.write("\ts: [0.."+str(self.nb_states-1)+"] init "+str(where(self.initial_state==1.0)[0][0])+";\n\n")
		self._savePrismMatrix(f)
		f.write('endmodule\n\n')

		labels = {}
		for s,l in enumerate(self.labelling):
			if l != 'init':
				if not l in labels:
					labels[l] = [str(s)]
				else:
					labels[l].append(str(s))
		for l in labels:
			res = 'label "'+l+'" ='
			for s in labels[l]:
				res += ' s='+s+' |'
			res = res[:-1]+';\n'
			f.write(res)
		f.close()


	def _save(self, f) -> None:
		f.write(str(self.labelling))
		f.write('\n')
		super()._save(f)

	def toStormpy(self):
		"""
		Returns the equivalent stormpy sparse model.
		The output object will be a stormpy.SparseDtmc.

		Returns
		-------
		stormpy.SparseDtmc
			The same model in stormpy format.
		"""
		try:
			from ..with_stormpy import jajapyModeltoStormpy
			return jajapyModeltoStormpy(self)
		except ModuleNotFoundError:
			raise RuntimeError("Stormpy is not installed on this machine.")

def labelsForRandomModel(nb_states: int, labelling: list) -> list:
	if 'init' in labelling:
		msg =  "The label 'init' cannot be used: it is reserved for initial states."
		raise SyntaxError(msg)

	if nb_states < len(labelling):
		print("WARNING: the size of the labelling is greater than the",end=" ")
		print("number of states. The last labels will not be assigned to",end=" ")
		print("any states.")
	

	if nb_states > len(labelling):
		print("WARNING: the size of the labelling is lower than the",end=" ")
		print("number of states. The labels for the last states will",end=" ")
		print("be chosen randomly.")

	labelling = labelling[:min(len(labelling),nb_states)] + choices(labelling,k=nb_states-len(labelling))
	labelling.append("init")
	return labelling



from ..base.tools import resolveRandom

class Scheduler:
	"""
	Abstract class for general Scheduler
	"""
	def __init__(self) -> None:
		pass
		
	def getAction(self) -> str:
		pass
		
	def reset(self) -> None:
		pass
	
	def addObservation(self,obs) -> None:
		pass

class UniformScheduler(Scheduler):
	"""
	Class for an uniform scheduler. An uniform scheduler will always returns
	each action with an equal probability.
	"""
	def __init__(self, actions: list) -> None:
		"""
		Creates a uniform scheduler for the given actions.

		Parameters
		----------
		actions : list of str
			A list of actions.
		"""
		super().__init__()
		self.actions = [[1/len(actions)]*len(actions),actions]
	
	def getAction(self) -> str:
		"""
		Returns the action chosen by the scheduler

		Returns
		-------
		str
			An action.
		"""
		return self.actions[1][resolveRandom(self.actions[0])]


class MemorylessScheduler(Scheduler):
	"""
	Class for memoryless scheduler.
	"""
	def __init__(self,actions: list) -> None:
		"""
		Creates a memoryless scheduler.

		Parameters
		----------
		actions : list of str
			list of ``n`` actions, with ``n`` the number of states in the model
			``actions[n]`` will always be used in state ``n``.
		"""
		super().__init__()
		self.actions = actions

	def getAction(self,current_state:int) -> str:
		"""
		Returns the action used while in state ``current_state``

		Parameters
		----------
		current_state : int
			ID of the current MDP state.

		Returns
		-------
		str
			Action chosen by the scheduler.
		"""
		return self.actions[current_state]

class FiniteMemoryScheduler:
	"""
	Class for finite memoryless scheduler.
	A finite memory scheduler can be seen as a generative automaton that takes
	on input a sequence of observation and returns a sequence of actions.
	"""
	def __init__(self,action_matrix: dict,transition_matrix: dict) -> None:
		"""
		Creates a finite memory scheduler.

		Parameters
		----------
		action_matrix : dict
			Describes the probability for each action to be chosen given the
			current scheduler state as follows:
			``action_matrix = {scheduler_state_x: [[proba1,proba2,...],[action1,action2,...]],
			scheduler_state_y: [[proba1,proba2,...],[action1,action2,...]],
			...}``

		transition_matrix : dict
			Given the a state ``s``and an observation ``o``, returns to which
			state the scheduler move to if it receives ``o`` while in state `s`
			``transition_matrix = {obs1: [scheduler_state_dest_if_current_state_=_0,scheduler_state_dest_if_current_state_=_1,...],
			obs2: [scheduler_state_dest_if_current_state_=_0,scheduler_state_dest_if_current_state_=_1,...]
			...}``
		"""
		super().__init__()
		self.s = 0
		self.action_matrix = action_matrix
		self.transition_matrix = transition_matrix

	def reset(self) -> None:
		"""
		Reset the model to its initial state. Should be done before generating
		a new trace.
		"""
		self.s = 0

	def getAction(self) -> str:
		"""
		Returns the action chosen by the scheduler

		Returns
		-------
		str
			Action chosen by the scheduler.
		"""
		return self.action_matrix[self.s][1][resolveRandom(self.action_matrix[self.s][0])]

	def addObservation(self,obs:str) -> None:
		"""
		Updates the scheduler state according to the current state and the
		given observation ``obs``.

		Parameters
		----------
		obs : str
			An observation.
		"""
		if obs in self.transition_matrix:
			self.s = self.transition_matrix[obs][self.s]

	def getActions(self,s:int=None) -> list:
		"""
		Returns the list of all the possibles actions (and their probabilities)
		that the scheduler can choose if it is in state ``s``. If ``s`` is not
		provided it considers the current scheduler state.

		Parameters
		----------
		s : int, optional
			A scheduler state ID. By default None

		Returns
		-------
		list
			A list containing a list of probabilities and a list of actions:
			[[proba_action1, proba_action2,...],[action1, action2,...]]
		"""
		if s==None:
			return self.action_matrix[self.s]
		else:
			return self.action_matrix[s]

	def getSequenceStates(self,seq_obs: list) -> list:
		"""
		Given a sequence of observations, returns the sequence of states the
		scheduler goes through.

		Parameters
		----------
		seq_obs : list of str
			A sequence of observations.

		Returns
		-------
		list
			A sequence of scheulder states.
		"""
		self.reset()
		res = [0]
		for o in seq_obs:
			self.add_observation(o)
			res.append(self.s)
		return res

	def getProbability(self,action:str,state:int=None) -> float:
		"""
		Given a scheduler state ID and an action, returns the probability to
		execute this action in this state.

		Parameters
		----------
		action : str
			An action.
		state : int, optional
			A state ID. If not provided it considers the current state.
			By default None

		Returns
		-------
		float
			A probability.
		"""
		if state == None:
			state = self.s
		if not action in self.action_matrix[state][1]:
			return 0
		return self.action_matrix[state][0][self.action_matrix[state][1].index(action)]



from .MC import MC
from ..base.Set import Set
from ..base.tools import hoeffdingBound
from numpy import zeros

class PTA_node:
	def __init__(self, label,parent=None,count=0) -> None:
		self.label = label
		self.count = count
		self.kids = []
		self.parent = parent
	
	def findKid(self,label):
		for (k,_) in self.kids:
			if k.label == label:
				return k
		return None

	def incKid(self,el,c=1):
		if type(el) == str:
			label = el
			s = -1
		else:
			s = el
			label = -1
		for i,(k,_) in enumerate(self.kids):
			if k.label == label or s==k:
				self.kids[i] = (k,self.kids[i][1]+c)
				return k
		return None

	def replaceKid(self,s1,s2):
		for i,(k,_) in enumerate(self.kids):
			if k == s1:
				self.kids[i] = (s2,self.kids[i][1])
				return True
		return None
	
class PTA:
	def __init__(self,traces) -> None:
		self.root = PTA_node('')
		self.alphabet = []
		
		for trace,time in zip(traces.sequences,traces.times):
			node = self.root
			curr = 0
			while curr < len(trace):
				char = trace[curr]
				if not char in self.alphabet:
					self.alphabet.append(char)
				nnext = node.incKid(char,time)
				if nnext == None:
					node = self.addKids(node,trace[curr:],time)
					curr = len(trace)
				else:
					node = nnext
					curr += 1
			node.count = time

	def addKids(self,start:PTA_node,string:str,count:int):
		n = start
		for l in string:
			n.kids.append((PTA_node(l,n),count))
			n = n.kids[-1][0]
		return n

	def compatible(self,s1,s2,alpha):
		if s1 == None or s2 == None:
			return True
		if s1.label != s2.label:
			return False
		n1 = sum([i for (_,i) in s1.kids])+s1.count
		n2 = sum([i for (_,i) in s2.kids])+s2.count
		if not hoeffdingBound(s1.count,n1,s2.count,n2,alpha):
			return False
		for char in self.alphabet:
			k1,k2 = s1.findKid(char),s2.findKid(char)
			if not self.compatible(k1,k2,alpha):
				return False
		return True
		
	def merge(self,s1,s2):
		s1.count += s2.count
		s2.parent.replaceKid(s2,s1)
		self.fold(s1,s2)

	def fold(self,s1,s2):
		for (k2,c) in s2.kids:
			k1 = s1.findKid(k2.label)
			if k1 == None:
				s1.kids.append((k2,c))
				k2.parent = s1
			else:
				k1.count += k2.count
				s1.incKid(k1,c)
				self.fold(k1,k2)

	def pprint(self):
		print(self.toMC())

	def toMC(self):
		root = self.root
		if 'init' in self.alphabet:
			if len(self.root.kids) == 1 and self.root.kids[0][0].label == 'init':
				root = self.root.kids[0][0]
			else:
				msg =  "The label 'init' cannot be used: it is reserved for initial states."
				raise SyntaxError(msg)
		else:
			root.label = 'init'
		
		states = [root]
		temp = [root]
		while len(temp) != 0:
			n = temp[0]
			temp = temp[1:]
			for (k,c) in n.kids:
				if not k in states:
					states.append(k)
					temp.append(k)
		
		matrix = zeros((len(states),len(states)))
		for i,s in enumerate(states):
			for (k,c) in s.kids:
				matrix[i,states.index(k)] = c
			if (matrix[i]>0).any():
				matrix[i] /= matrix[i].sum()
		labelling = [s.label for s in states]
		return MC(matrix,labelling)
		

class Alergia:
	"""
	Class for general ALERGIA algorithm described here:
	https://grfia.dlsi.ua.es/repositori/grfia/pubs/57/icgi1994.pdf
	"""

	def __init__(self):
		None

	def fit(self,traces: Set,alpha: float=0.1,
			stormpy_output: bool = True, output_file_prism: str = None):
		"""
		Fits a MC according to ``traces``.

		Parameters
		----------
		traces : Set
			The training set.
		alpha : float, optional
			_description_, by default 0.1
		stormpy_output: bool, optional
			If set to True the output model will be a Stormpy sparse model.
			Default is True.
		output_file_prism : str, optional
			If set, the output model will be saved in a prism file at this
			location. Otherwise the output model will not be saved.

		Returns
		-------
		MC or stormpy.SparseDtmc
			The fitted MC.
			If `stormpy_output` is set to `False` or if stormpy is not available on
			the machine it returns a `jajapy.MC`, otherwise it returns a `stormpy.SparseDtmc`
		"""
		self._initialize(traces,alpha)
		red = [self.T.root]
		blue = [k for (k,_) in self.T.root.kids]
		while len(blue)>0:
			s1 = blue[0]
			merged = False
			for s2 in red:
				if self.T.compatible(s1,s2,alpha):
					self.T.merge(s2,s1)
					merged = True
					break
			if not merged:
				red.append(s1)
			blue = []
			for s in red:
				blue += [k for (k,_) in s.kids]
			blue = list(set(blue))
			blue = [s for s in blue if s not in red]
		m = self.T.toMC()

		try:
			from ..with_stormpy import jajapyModeltoStormpy
			stormpy_installed = True
		except ModuleNotFoundError:
			stormpy_installed = False
		if stormpy_output and not stormpy_installed:
			print("WARNING: stormpy not found. The output model will not be a stormpy sparse model")
			stormpy_output = False
		
		if output_file_prism:
			self.h.savePrism(output_file_prism)

		if stormpy_output:
			return jajapyModeltoStormpy(m)
		else:
			return m

	def _initialize(self,traces:Set,alpha:float) -> None:
		"""
		Create the PTA from the training set ``traces ``, initialize
		the alpha value and the alphabet.

		Parameters
		----------
		traces : Set
			Training set.
		alpha : float
			alpha value used in the Hoeffding bound
		"""
		self.alpha = alpha
		self.alphabet = traces.getAlphabet()
		self.createPTA(traces)

	
	def createPTA(self,traces):
		temp1 = traces.sequences[:]
		temp2 = traces.times[:]
		traces.sequences.sort()
		traces.times = [temp2[temp1.index(i)] for i in traces.sequences]
		self.T = PTA(traces)
		
		



import unittest
from ..hmm import *
from ..mc import *
from ..ctmc import *
from ..pctmc import *
from ..mdp import *
from ..with_stormpy import jajapyModeltoStormpy, stormpyModeltoJajapy, loadPrism
import stormpy
from os import remove

def modelMDP(p=0.75):
	labelling = ['R','L','R','L','OK','HIT']
	transitions = [(0,'m',1,p),(0,'m',2,1-p),(0,'s',3,p),(0,'s',0,1-p),
				   (1,'m',0,p),(1,'m',3,1-p),(1,'s',2,p),(1,'s',1,1-p),
				   (2,'m',5,1.0),(2,'s',4,1.0),(3,'m',5,1.0),(3,'s',4,1.0),
				   (4,'m',4,1.0),(4,'s',4,1.0),(5,'m',5,1.0),(5,'s',5,1.0)]
	return createMDP(transitions,labelling,0,"bigstreet")

def modelMC():
	labelling = list("BTSXSPTXPVVE")
	initial_state = 0
	name = "MC_REBER"
	transitions = [(0,1,0.5),(0,5,0.5),(1,2,0.6),(1,3,0.4),(2,2,0.6),(2,3,0.4),
				   (3,7,0.5),(3,4,0.5),(4,11,1.0),(5,6,0.7),(5,9,0.3),
				   (6,6,0.7),(6,9,0.3),(7,6,0.7),(7,9,0.3),(8,7,0.5),(8,4,0.5),
				   (9,8,0.5),(9,10,0.5),(10,11,1.0),(11,11,1.0)]
	return createMC(transitions,labelling,initial_state,name)

def modelCTMC():
	labelling = ['red','red','yellow','blue','blue']
	transitions = [(0,1,0.08),(0,2,0.12),(1,1,0.3),(1,2,0.7),
				   (2,0,0.2),(2,3,0.1),(2,4,0.2),(3,3,0.8),
				   (3,1,0.1),(3,4,0.1),(4,2,0.25)]
	return createCTMC(transitions,labelling,0,"My_CTMC")

def modelPCTMC():
	labelling = ['c_red_p_red','c_green_p_red','c_orange_p_green','c_red_p_green','c_orange_p_red']
	transitions = [(0,1,5.0),(1,2,'p*p/2'),(2,3,2.0),(2,4,0.1),(3,0,0.1),(4,0,2.0)]
	parameters = ['p']
	return createPCTMC(transitions,labelling,parameters,0)


class ModelConverterTestclass(unittest.TestCase):

	def test_MC_Stormpy(var):
		m = modelMC()
		mstorm = jajapyModeltoStormpy(m)
		properties = stormpy.parse_properties('P=? [F "E"]')
		result = stormpy.check_model_sparse(mstorm,properties[0])
		var.assertAlmostEqual(result.at(mstorm.initial_states[0]),1.0)

		properties = stormpy.parse_properties('P=? [G !"S"]')
		result = stormpy.check_model_sparse(mstorm,properties[0])
		var.assertAlmostEqual(result.at(mstorm.initial_states[0]),0.4)

		mprime = stormpyModeltoJajapy(mstorm)
		mprime.name = m.name
		var.assertEqual(str(mprime),str(m))
	
	def test_MC_Prism(var):
		m = modelMC()
		ts = m.generateSet(1000,15)
		mprime = loadPrism("jajapy/tests/materials/mc/reber.sm")
		var.assertAlmostEqual(m.logLikelihood(ts),mprime.logLikelihood(ts))

		m.savePrism("test_save.sm")
		mprime = loadPrism("test_save.sm")
		var.assertAlmostEqual(m.logLikelihood(ts),mprime.logLikelihood(ts))
		remove("test_save.sm")
	
	def test_MDP_Stormpy(var):
		m = modelMDP()
		actions = m.actions
		#actions.reverse()
		mstorm = jajapyModeltoStormpy(m)
		mprime = stormpyModeltoJajapy(mstorm,actions)
		mprime.name = m.name
		var.assertEqual(str(mprime),str(m))
	
	def test_MDP_Prism(var):
		m = modelMDP()
		ts = m.generateSet(1000,15,scheduler=UniformScheduler(m.actions))
		mprime = loadPrism("jajapy/tests/materials/mdp/bigstreet.sm")
		mprime.actions = m.actions #sometimes we need to reverse it
		var.assertEqual(m.logLikelihood(ts),mprime.logLikelihood(ts))

		m.savePrism("test_save.sm")
		mprime = loadPrism("test_save.sm")
		mprime.actions = m.actions #sometimes we need to reverse it
		var.assertEqual(m.logLikelihood(ts),mprime.logLikelihood(ts))
		remove("test_save.sm")

	def test_CTMC_Stormpy(var):
		m = modelCTMC()
		ts = m.generateSet(1000,15,timed=True)
		mprime = jajapyModeltoStormpy(m)
		mprime = stormpyModeltoJajapy(mprime)
		var.assertEqual(m.logLikelihood(ts),mprime.logLikelihood(ts))
	
	def test_CTMC_Prism(var):
		m = modelCTMC()
		ts = m.generateSet(1000,15,timed=True)
		mprime = loadPrism("jajapy/tests/materials/ctmc/model_ctmc_prism.sm")
		var.assertEqual(m.logLikelihood(ts),mprime.logLikelihood(ts))

		m.savePrism("test_save.sm")
		mprime = loadPrism("test_save.sm")
		var.assertEqual(m.logLikelihood(ts),mprime.logLikelihood(ts))
		remove("test_save.sm")
	
	def test_PCTMC_Stormpy(var):
		m = modelPCTMC()
		m.instantiate(['p'],[2.0])
		mprime = jajapyModeltoStormpy(m)
		mprime = stormpyModeltoJajapy(mprime)
		ts = m.generateSet(1000,15,timed=True)
		var.assertEqual(m.logLikelihood(ts),mprime.logLikelihood(ts))
	
	def test_PCTMC_Prism(var):
		m = modelPCTMC()
		m.savePrism("test_save.sm")
		m.instantiate(['p'],[2.0])
		ts = m.generateSet(1000,15,timed=True)
		mprime = loadPrism("jajapy/tests/materials/pctmc/tl.pm")
		mprime.instantiate(['p'],[2.0])
		var.assertEqual(m.logLikelihood(ts),mprime.logLikelihood(ts))
		
		mprime = loadPrism("test_save.sm")
		mprime.instantiate(['p'],[2.0])
		var.assertEqual(m.logLikelihood(ts),mprime.logLikelihood(ts))
		remove("test_save.sm")


if __name__ == "__main__":
	unittest.main()


from numpy import ndarray, array, isnan, where, nan
from numpy.random import geometric
from ast import literal_eval
from copy import deepcopy
from ..base.Set import Set
from ..base.Model import Model, PCTMC_ID
from sympy import sympify

class Parametric_Model(Model):
	"""
	Abstract class that represents parametric models.
	Is inherited by PCTMC.
	Should not be instantiated itself!
	"""
	def __init__(self, matrix: ndarray, labelling: list,
				 transition_expr: list, parameter_values: dict,
				 parameter_indexes: list, parameter_str: list,
				 name: str) -> None:
		"""
		Creates an abstract model for PMC or PCTMC.

		Parameters
		----------
		matrix : ndarray
			Represents the transition matrix.
			matrix[i,j] is the index, in `transition_expr`, of the symbolic
			representation of the transition from `i` to `j`.
		labelling: list of str
			A list of N observations (with N the nb of states).
			If `labelling[s] == o` then state of ID `s` is labelled by `o`.
			Each state has exactly one label.
		transition_expr: list of str
			Contains the symbolic representation for each transition.
		parameter_values: list of float
			Contains the value for each parameter.
			`parameter_values[i]` is the instantiation for parameter `i`.
			If the ith parameter is not instantiated, `parameter_values[i] == numpy.nan`.
		parameter_indexes: list of ndarray
			Contains the indexes of each transition using each parameter.
			`parameter_indexes[i] = array([[0,1],[2,1]])` means that parameter `i`
			is used by the transition from state 0 to state 1 and from state 2 to state 1.
		parameter_str: list
			Contains the name of each parameter.
			`parameter_str[i]` is the name of parameter `i`.
			Parameter `i` doesn't have a name if `parameter_str[i]==None`.
		name : str, optional
			Name of the model.
		"""
		self.labelling = labelling
		self.alphabet = list(set(labelling))
		self.nb_states = len(matrix)
		
		if not 'init' in self.labelling:
			msg = "No initial state given: at least one"
			msg += " state should be labelled by 'init'."
			raise ValueError(msg)
		initial_state = [1.0/self.labelling.count("init") if i=='init' else 0.0 for i in self.labelling]

		if len(parameter_indexes) != len(parameter_str):
			raise ValueError("Length of parameter_indexes and parameter_str must be equal.")
		self.nb_parameters = len(parameter_indexes)
		if max(matrix.flatten()) >= len(transition_expr):
			msg = "Transition "+str(max(matrix.flatten()))+" found in matrix while length "
			msg+= "of transition_expr is "+str(len(transition_expr))
			raise ValueError(msg)
		if min(matrix.flatten()) < 0:
			raise ValueError("Transition "+str(min(matrix.flatten()))+" found in matrix")
		
		if len(self.labelling) != self.nb_states:
			msg = "The length of labelling ("+str(len(labelling))+") is not equal "
			msg+= "to the number of states("+str(self.nb_states)+")"
			raise ValueError(msg)
		
		self.transition_expr = transition_expr
		k = list(parameter_values.keys())
		for i in k:
			if isnan(parameter_values[i]):
				del parameter_values[i]
		self.parameter_values = parameter_values
		self.parameter_indexes= parameter_indexes
		self.parameter_str = parameter_str

		super().__init__(matrix,initial_state,name)

	def isInstantiated(self,state1:int = None, state2:int = None, param:str =None) -> bool:
		"""
		If nothing is set, checks if all the parameters are instantiated.
		If `state1` is set, checks if all the transitions leaving this state
		are instantiated.
		If `state1` and `state2` are set, checks if the transitions from
		`state1` to `state2` is instantiated.
		If `param` is set, checks if the parameter `param` is instantiated.

		Parameters
		----------
		state1 : int, optional.
			state ID.
		
		state2 : int, optional.
			state ID.
		
		param : str, optional
			Parameter name.
			
		Returns
		-------
		bool
			True if all the requested entity(ies) is/are instantiated.
			See above.
		"""
		if type(param) == str:
			return param in self.parameter_values
		elif type(state1) == type(None):
			return len(set(self.parameter_str)-set(self.parameter_values.keys())) == 0
		else:
			self._checkStateIndex(state1)
			if type(state2) == type(None):
				parameters = self.involvedParameters(state1)
			else:
				parameters = self.involvedParameters(state1,state2)
			for i in parameters:
				if not self.isInstantiated(param=i):
					return False
			return True

	def transitionValue(self,i:int,j:int):
		"""
		Returns the value of the transition from state `i` to `j`.
		If at least one of the paramater in this transition is not
		instantiated, returns the symbolic representation of the transition.

		Parameters
		----------
		i : int
			source state ID.
		j : int
			destination state ID.

		Returns
		-------
		float or symbolic representation
			The value of the transition from state `i` to `j`.
			If at least one of the paramater in this transition is not
			instantiated, returns the symbolic representation of the
			transition.
		"""
		return self.transition_expr[self.matrix[i,j]].evalf(subs=self.parameter_values)

	def transitionExpression(self,i:int,j:int):
		"""
		Returns the symbolic representation of the transition.

		Parameters
		----------
		i : int
			source state ID.
		j : int
			destination state ID.

		Returns
		-------
		symbolic representation
			The symbolic representation of the transition.
		"""
		return self.transition_expr[self.matrix[i,j]]
	
	def parameterValue(self, p:str)-> float:
		"""
		Returns the value of the parameter `p`.
		If `p` is not instantiated, returns numpy.nan.
		If the model doesn't have any parameter `p`, returns numpy.nan.

		Parameters
		----------
		p : str
			parameter name.

		Returns
		-------
		float
			The value of `p` if `p` is instantiated, numpy.nan otherwise.
			If the model doesn't have any parameter `p`, returns numpy.nan.
		"""
		if not p in self.parameter_values:
			return nan
		return self.parameter_values[p]
	
	def parameterIndexes(self, p:str) -> list:
		"""
		Returns the list of all transitions involving `p`.
		If the model doesn't have any parameter `p`, returns an empty list.

		Parameters
		----------
		p : str
			parameter name.

		Returns
		-------
		list
			The list of all transitions involving `p`.
			If the model doesn't have any parameter `p`, returns an empty list.
		
		Examples
		--------
		>>> m.parameterIndexes('x')
		[[4,0],[2,3]]
		>>> # the transitions from state 4 to 0 and from 2 to 3 involve 'x'.
		>>> m.transitionExpression(4,0)
		3*x
		"""
		if not p in self.parameter_str:
			return []
		return self.parameter_indexes[self.parameter_str.index(p)]

	def instantiate(self,parameters: list, values: list) -> ndarray: 
		"""
		Returns a copy of the `self.parameter_values` dict, where all the
		parameters in `parameters` to the values `values`.
		
		Parameters
		----------
		parameters : list of string
			List of all parameters to set. This list must contain parameter
			names.
		values : list of float
			List of values. `parameters[i]` will be set to `values[i]`.

		Returns
		-------
		numpy.ndarray
			A copy of the `self.parameter_values` dict, where all the
			parameters in `parameters` to the values `values`.
		
		Remark
		------
		The output numpy.ndarray will be used by the inherited class to check
		if this instantiation is valid. If so, the `self.parameter_values` dict
		is updated.
		"""
		new_values = deepcopy(self.parameter_values)
		for s,v in zip(parameters,values):
			new_values[s] = v
		return new_values

	#def evaluateString(self,string:str,parameter_values=None):
	#	if parameter_values == None:
	#		parameter_values = self.parameter_values
	#	return sympify(string).evalf(subs=parameter_values)

	def evaluateTransition(self,i:int,j:int, parameter_values: dict = None):
		"""
		Returns the value of the transition from state `i` to `j`.
		If `parameter_values` is given, the parameters are instantiated by
		`parameter_values`, otherwise by `self.parameter_values`.
		If at least one of the paramater in this transition is not
		instantiated, returns the symbolic representation of the transition.

		Parameters
		----------
		i : int
			source state ID.
		j : int
			destination state ID.
		parameter_values: dict, optional
			dictionary with the instantiation of the parameters.
			The keys are the parameter names (str).

		Returns
		-------
		float or symbolic representation
			The value of the transition from state `i` to `j`.
			If at least one of the paramater this transition is not
			instantiated, returns the symbolic representation of the
			transition.
		"""
		if parameter_values == None:
			parameter_values = self.parameter_values
		v =  self.transitionExpression(i,j).evalf(subs=parameter_values)
		try:
			v = float(v)
		except TypeError:
			pass
		return v

	def involvedParameters(self,i: int,j: int = -1) -> list:
		"""
		Returns the parameters involved in the transition from `i` to `j`.
		If `j` is not set, it returns the parameters involved in all the 
		transitions leaving `i`.

		Parameters
		----------
		i : int
			source state ID.
		j : int, optional
			destination state ID.

		Returns
		-------
		list of str
			list of parameter names.
		"""
		if j == -1:
			j = range(self.nb_states)
		else:
			j =  [j]
		res = set()
		for jj in j:
			res = res.union(self.transitionExpression(i,jj).free_symbols)
		res = list(res)
		res = [i.name for i in res]
		return res

	def getLabel(self,state: int) -> str:
		"""
		Returns the label of `state`.

		Parameters
		----------
		state : int
			a state ID

		Returns
		-------
		str
			a label

		Example
		-------
		>>> model.getLabel(2)
		'Label-of-state-2'
		"""
		self._checkStateIndex(state)
		return self.labelling[state]
	
	def getAlphabet(self) -> list:
		"""
		Returns the alphabet of this model.

		Returns
		-------
		list of str
			The alphabet of this model
		
		Example
		-------
		>>> model.getAlphabet()
		['a','b','c','d','done']
		"""
		return self.alphabet
	
	def __str__(self) -> str:
		res = "Name: "+self.name+'\n'
		for i in self.parameter_values.keys():
			if not isnan(self.parameter_values[i]):
				res += i+' = '+str(round(self.parameter_values[i],5))+'\n'
		res += '\n'
		if 1.0 in self.initial_state:
			res += "Initial state: s"+str(where(self.initial_state==1.0)[0][0])+'\n'
		else:
			res += "Initial state: "
			for i in range(len(self.matrix)):
				if self.initial_state[i] >= 0.001:
					res += 's'+str(i)+': '+str(round(self.initial_state[i],3))+', '
			res = res[:-2]+'\n'
		for i in range(self.nb_states):
			res += self._stateToString(i)+'\n'
		res += '\n'
		return res

	def generateSet(self, set_size: int, param, distribution=None, min_size=None, timed: bool=False) -> Set:
		"""
		Generates a set (training set / test set) containing ``set_size`` traces.

		Parameters
		----------
		set_size: int
			number of traces in the output set.
		param: a list, an int or a float.
			the parameter(s) for the distribution. See "distribution".
		distribution: str, optional
			If ``distribution=='geo'`` then the sequence length will be
			distributed by a geometric law such that the expected length is
			``min_size+(1/param)``.
			If distribution==None param can be an int, in this case all the
			seq will have the same length (``param``), or ``param`` can be a
			list of int.
			Default is None.
		min_size: int, optional
			see "distribution". Default is None.
		timed: bool, optional
			Only for timed model. Generate timed or non-timed traces.
			Default is False.
		
		Returns
		-------
		output: Set
			a set (training set / test set).
		
		Examples
		--------
		>>> set1 = model.generateSet(100,10)
		>>> # set1 contains 100 traces of length 10
		>>> set2 = model.generate(100, 1/4, "geo", min_size=6)
		>>> # set2 contains 100 traces. The length of the traces is distributed following
		>>> # a geometric distribution with parameter 1/4. All the traces contains at
		>>> # least 6 observations, hence the average length of a trace is 6+(1/4)**(-1) = 10.
		"""
		seq = []
		val = []
		for i in range(set_size):
			if distribution == 'geo':
				curr_size = min_size + int(geometric(param))
			else:
				if type(param) == list:
					curr_size = param[i]
				elif type(param) == int:
					curr_size = param

			if timed:
				trace = self.run(curr_size,timed)
			else:
				trace = self.run(curr_size)

			if not trace in seq:
				seq.append(trace)
				val.append(1)
			else:
				val[seq.index(trace)] += 1

		return Set(seq,val)

	def _save(self, f) -> None:
		f.write(self.name)
		f.write('\n')
		f.write(str(self.matrix.tolist()))
		f.write('\n')
		f.write(str(self.labelling))
		f.write('\n')
		f.write(str(self.parameter_values))
		f.write('\n')
		f.write(str(self.parameter_indexes))
		f.write('\n')
		f.write(str(self.parameter_str))
		f.write('\n')
		f.write(str([str(i) for i in self.transition_expr]))
		f.write('\n')
		f.close()

	def _checkStateIndex(self,s:int) -> None:
		try:
			s = int(s)
		except TypeError:
			raise TypeError('The parameter must be a valid state ID')
		if s < 0:
			raise IndexError('The parameter must be a valid state ID')
		elif s >= self.nb_states:
			raise IndexError('This model contains only '+str(self.nb_states)+' states')

def loadParametricModel(f):
	"""
	Used by the inherited classes.
	Should not be used directly!
	"""
	name = f.readline()[:-1]
	matrix = array(literal_eval(f.readline()[:-1]))
	labelling = literal_eval(f.readline()[:-1])
	parameter_values = literal_eval(f.readline()[:-1])
	parameter_indexes = literal_eval(f.readline()[:-1])
	parameter_str = literal_eval(f.readline()[:-1])
	transition_expr = literal_eval(f.readline()[:-1])
	transition_expr = [sympify(i) for i in transition_expr]
	return matrix,labelling,parameter_values,parameter_indexes,parameter_str,transition_expr,name


from ..base.tools import resolveRandom
from ..base.Parametric_Model import *
from ..base.Set import Set
from numpy import ndarray, array, zeros, dot, newaxis, hstack, inf, vstack, delete
from numpy.random import exponential, rand
from math import exp, log
from multiprocessing import cpu_count, Pool
from sys import platform
from sympy import sympify

class PCTMC(Parametric_Model):
	"""
	Class representing a PCTMC.
	"""
	def __init__(self, matrix: ndarray, labelling: list,
				 transition_expr: list, parameter_values: dict,
				 parameter_indexes: list, parameter_str: list,
				 name: str="unknow_PCTMC",
				 synchronous_transitions: list = []) -> None:
		"""
		Creates an PCTMC.

		Parameters
		----------
		matrix : ndarray
			Represents the transition matrix.
			matrix[i,j] is the index, in `transition_str`, of the symbolic
			value of the transition from `i` to `j`.
		labelling: list of str
			A list of N observations (with N the nb of states).
			If `labelling[s] == o` then state of ID `s` is labelled by `o`.
			Each state has exactly one label.
		transition_str: list of str
			Contains the symbolic value for each transition.
		parameter_values: dict
			Contains the value for each instantiated parameter.
			`parameter_values[i]` is the instantiation for parameter `i`.
			If the ith parameter is not instantiated, `parameter_values[i] == nan`.
		parameter_indexes: list of ndarray
			Contains the indexes of each transition using each parameter.
			`parameter_indexes[i] = array([[0,1],[2,1]])` means that parameter `i`
			is used by the transition from state 0 to state 1 and from state 2 to state 1.
		parameter_str: list
			Contains the name of each parameter.
			`parameter_str[i]` is the name of parameter `i`.
			Parameter `i` doesn't have a name if `parameter_str[i]==None`.
		name : str, optional
			Name of the model.
			Default is "unknow_PCTMC"
		synchronous_transitions: list, optional.
			This is useful only for synchronously composing this PCTMC with
			another one.
			List of (source_state <int>, action <str>, dest_state <int>, rate <float>).
			Default is an empty list.
		"""
		self.model_type = PCTMC_ID
		self.synchronous_transitions = synchronous_transitions
		super().__init__(matrix,labelling,transition_expr,parameter_values,parameter_indexes,parameter_str,name)

	def e(self,s: int) -> float:
		"""
		Returns the exit rate of state ``s``, i.e. the sum of all the rates in
		this state.

		Returns
		-------
		s : int
			A state ID.
		float
			An exit rate.
		"""
		if self.isInstantiated(s):
			return sum([self.evaluateTransition(s,i) for i in range(self.nb_states)])
		raise ValueError("cannot compute the exit rate of non-instantiated state "+str(s))

	def l(self, s1:int, s2:int, obs:str) -> float:
		"""
		Returns the rate associated to the transition from state `s1`, seeing
		`obs`, to state `s2`.

		Parameters
		----------
		s1 : int
			A state ID.
		s2 : int
			A state ID.
		obs : str
			An observation.

		Returns
		-------
		float
			A rate.
		"""
		self._checkStateIndex(s1)
		self._checkStateIndex(s2)
		if self.labelling[s1] != obs:
			return 0.0
		return self.evaluateTransition(s1,s2)
	
	def lkl(self, s: int, t: float) -> float:
		"""
		Returns the likelihood of staying in `s` state for a duration `t`.

		Parameters
		----------
		s : int
			A state ID.
		t : float
			A waiting time.

		Returns
		-------
		float
			A Likelihood.
		"""
		if t < 0.0:
			return 0.0
		try:
			return self.e(s)*exp(-self.e(s)*t)
		except ValueError:
			print("WARN: cannot compute the lkl of non-instantiated state "+str(s))

	def tau(self, s1:int, s2: int, obs: str) -> float:
		"""
		Returns the probability to move from `s1` to `s2` generating
		observation ``obs``.

		Parameters
		----------
		s1 : int
			A state ID.
		s2 : int
			A state ID.
		obs : str
			An observation.

		Returns
		-------
		float
			A probability.
		"""
		try:
			e = self.e(s1)
			if e == 0.0:
				return inf
			return self.l(s1,s2, obs)/e
		except ValueError:
			print("WARN: cannot compute the exit-rate of non-instantiated state "+str(s1))
	
	def expected_time(self, s:int) -> float:
		"""
		Returns the expected waiting time in `s`, i.e. the inverse of
		the exit rate.
		
		Parameters
		----------
		s : int
			A state ID.
		
		Returns
		-------
		float
			expected waiting time in this state.
		"""
		try:
			e = self.e(s)
			if e == 0.0:
				return inf
			return 1/e
		except ValueError:
			print("WARN: cannot compute the expected time of non-instantiated state "+str(s))

	def _checkRates(self,values=None) -> bool:
		if type(values) == type(None):
			values = self.parameter_values
		xs,ys = where(self.matrix != 0)
		for x,y in zip(xs,ys):
			v = self.evaluateTransition(x,y)
			if type(v) == float:
				if v < 0.0:
					return False
		return True
		
	def instantiate(self, parameters: list, values: list) -> bool:
		"""
		Set all the parameters in `parameters` to the values `values`.

		Parameters
		----------
		parameters : list of string
			List of all parameters to set. This list must contain parameter
			names.
		values : list of float
			List of values. `parameters[i]` will be set to `values[i]`.
		
		Returns
		-------
		bool
			True if the instantiation is valid (no transition has a negative
			evaluation) and then applied, False if the instantiation is not 
			valid (and then ignored).
		"""
		new_values =  super().instantiate(parameters, values)
		if self._checkRates(new_values):
			self.parameter_values = new_values
			return True
		return False

	def randomInstantiation(self, parameters: list = None,min_val:float = None,max_val:float = None) -> None:
		"""
		Randomly instantiated the parameters given in `parameters`.
		If `parameters` is not set it instantiates all the non-instantiated
		parameters in the model.

		Parameters
		----------
		parameters : list of string, optional.
			List of parameters names.
		min_val : float, optional
			Minimal value for the randomly instantiated parameters.
			If not set and if the model has at least two instantiated parameters,
			this value is equal to the parameters with the smallest instantiation.
			If not set and if the model has less than two instantiated parameters,
			this value is equal to 0.1.
		max_val : float, optional
			Maximal value for the randomly instantiated parameters.
			If not set and if the model has at least two instantiated parameters,
			this value is equal to the parameters with the highest instantiation.
			If not set and if the model has less than two instantiated parameters,
			this value is equal to 5.0.
		"""
		if parameters == None:
			parameters = []
			for i in self.parameter_str:
				if not i in self.parameter_values:
					parameters.append(i)
				elif isnan(self.parameter_values[i]):
					parameters.append(i)
		if min_val == None:
			if len(list(self.parameter_values.values())) > 1:
				min_val = min(self.parameter_values.values())
			else:
				min_val = 0.01
		if max_val == None:
			if len(list(self.parameter_values.values())) > 1:
				max_val = max(self.parameter_values.values())
			else:
				max_val = 0.3
		val = rand(len(parameters))*(max_val-min_val)+min_val
		while not self.instantiate(parameters,val):
			val = rand(len(parameters))*(max_val-min_val)+min_val
	
	def _stateToString(self,state:int) -> str:
		res = "----STATE "+str(state)+"--"+self.labelling[state]+"----\n"
		if self.isInstantiated(state):
			res += "Exepected waiting time: "+str(round(self.expected_time(state),5))+'\n'
		for j in range(len(self.matrix[state])):
			if self.matrix[state][j] != 0:
				val = str(self.transitionExpression(state,j))
				if self.isInstantiated(state,j) and len(self.involvedParameters(state,j)) > 0:
					val += ' (='+str(round(self.transitionValue(state,j),5))+')'
				res += "s"+str(state)+" -> s"+str(j)+" : lambda = "+val+'\n'
		return res
	
	def next(self,state: int) -> tuple:
		"""
		Return a state-observation pair according to the distributions 
		described by matrix
		
		Returns
		-------
		output : (int, str)
			A state-observation pair.
		"""
		if not self.isInstantiated(state):
			raise ValueError("At least one of the parameter is not instantiated.")
		exps = []
		for exp_lambda in [self.transitionValue(state,j) for j in range(self.nb_states)]:
			if exp_lambda == 0.0:
				exps.append(inf)
			else:
				exps.append(exponential(1/exp_lambda))
		next_state= exps.index(min(exps))
		next_obs  = self.labelling[state]
		return (next_obs, next_state, min(exps))

	def run(self,number_steps: int, timed: bool = False) -> list:
		"""
		Simulates a run of length ``number_steps`` of the model and return the
		sequence of observations generated. If ``timed`` it returns a list of
		pairs waiting time-observation.
		
		Parameters
		----------
		number_steps: int
			length of the simulation.
		timed: bool, optional
			Wether or not it returns also the waiting times. Default is False.

		Returns
		-------
		output: list of str
			trace generated by the run.
		"""
		output = []
		current = resolveRandom(self.initial_state)
		c = 0
		while c < number_steps:
			[symbol, next_state, time_spent] = self.next(current)
			output.append(symbol)
			if timed:
				output.append(time_spent)
			current = next_state
			c += 1
		output.append(self.labelling[current])
		return output
	
	def _computeAlphas_timed(self, sequence: list, times: int) -> float:
		obs_seq   = [sequence[i] for i in range(0,len(sequence),2)]
		times_seq = [sequence[i] for i in range(1,len(sequence),2)]
		len_seq = len(obs_seq)-1
		prev_arr = array(self.initial_state)
		for k in range(len_seq):
			new_arr = zeros(self.nb_states)
			for s in range(self.nb_states):
				p = array([self.l(ss,s,obs_seq[k])*exp(-self.e(ss)*times_seq[k]) for ss in range(self.nb_states)])
				new_arr[s] = dot(prev_arr,p)
			prev_arr = new_arr
		last_arr = prev_arr * (array(self.labelling) == obs_seq[-1])
		return log(last_arr.sum())*times

	def logLikelihood(self,traces: Set) -> float:
		"""
		Computes the loglikelihood of `traces` under this model.

		Parameters
		----------
		traces : Set
			a set of traces.

		Returns
		-------
		float
			the loglikelihood of `traces` under this model.
		"""
		if not self.isInstantiated():
			raise ValueError("Cannot compute the loglikelihood of a set under a non-instantiated model.")
			
		if traces.type == 0: # non-timed traces
			return super().logLikelihood(traces)
		else: # timed traces
			if platform != "win32" and platform != "darwin":
				p = Pool(processes = cpu_count()-1)
				tasks = []
				for seq,times in zip(traces.sequences,traces.times):
					tasks.append(p.apply_async(self._computeAlphas_timed, [seq, times,]))
				temp = [res.get() for res in tasks if res.get() != False]
			else:
				temp = [self._computeAlphas_timed(seq,times) for seq,times in zip(traces.sequences,traces.times)]
			return sum(temp)/sum(traces.times)

	def save(self,file_path:str) -> None:
		"""Save the model into a text file.

		Parameters
		----------
		file_path : str
			path of the output file.
		
		Examples
		--------
		>>> model.save("my_model.txt")
		"""
		f = open(file_path, 'w')
		f.write("PCTMC\n")
		super()._save(f)
	
	def savePrism(self,file_path:str) -> None:
		"""
		Save this model into `file_path` in the Prism format.

		Parameters
		----------
		file_path : str
			Path of the output file.
		"""
		f = open(file_path,'w')
		f.write("ctmc\n\n")

		for p in self.parameter_str:
			f.write("const double "+p)
			if self.isInstantiated(param=p):
				f.write(' = '+str(self.parameterValue(p)))
			f.write(';\n')

		f.write("\nmodule "+self.name+'\n')
		f.write("\ts: [0.."+str(self.nb_states)+"] init "+str(where(self.initial_state==1.0)[0][0])+";\n\n")
		
		for s1 in range(self.nb_states):
			if (self.matrix[s1]!=0).any():
				res = '\t[] s='+str(s1)+' ->'
				for s2 in where(self.matrix[s1] != 0)[0]:
					t = str(self.transition_expr[self.matrix[s1,s2]])
					t = t.replace('**','^')
					res += ' '+t+" : (s'="+str(s2)+") +"
				res = res[:-2]+';\n'
				f.write(res)
		f.write('endmodule\n\n')

		labels = {}
		for s,l in enumerate(self.labelling):
			if l != 'init':
				if not l in labels:
					labels[l] = [str(s)]
				else:
					labels[l].append(str(s))
		for l in labels:
			res = 'label "'+l+'" ='
			for s in labels[l]:
				res += ' s='+s+' |'
			res = res[:-1]+';\n'
			f.write(res)
		f.close()
		
	
def loadPCTMC(file_path: str) -> PCTMC:
	"""
	Load an PCTMC saved into a text file.

	Parameters
	----------
	file_path : str
		Location of the text file.
	
	Returns
	-------
	output : PCTMC
		The PCTMC saved in `file_path`.
	
	Examples
	--------
	>>> model = loadPCTMC("my_model.txt")
	"""
	f = open(file_path,'r')
	l = f.readline()[:-1] 
	if l != "PCTMC":
		msg = "This file doesn't describe a PCTMC: it describes a "+l
		raise ValueError(msg)
	matrix,labelling,parameter_values,parameter_indexes,parameter_str,transition_expr,name = loadParametricModel(f)
	f.close()
	return PCTMC(matrix,labelling,transition_expr,parameter_values,
				 parameter_indexes,parameter_str,name)

def createPCTMC(transitions: list, labelling: list, parameters: list,
				initial_state, parameter_instantiation: dict={},
				synchronous_transitions=[], name: str ="unknown_PCTMC") -> PCTMC:
	"""
	An user-friendly way to create a PCTMC.

	Parameters
	----------
	transitions : [ list of tuples (int, int, float or str)]
		Each tuple represents a transition as follow: 
		(source state ID, destination state ID, probability).
		The probability can be explicitly given (then it's a float),
		or a parameter (then it's the name of the parameter).
	labelling: list of str
		A list of N observations (with N the nb of states).
		If `labelling[s] == o` then state of ID `s` is labelled by `o`.
		Each state has exactly one label.
	parameters: list of str.
		A list containing all the parameters name.
	initial_state : int or list of float
		Determine which state is the initial one (then it's the id of the
		state), or what are the probability to start in each state (then it's
		a list of probabilities).
	parameter_instantiation: dict
		An instantion for some (or all) parameters.
		`parameter_instantiation == {'p':0.5}` means that parameter `p`
		should be instantiated to 0.5. The other parameters are not
		instantiated.
	synchronous_transitions: list, optional.
		This is useful only for synchronously composing this PCTMC with
		another one.
		List of (source_state <int>, action <str>, dest_state <int>, rate <float>).
		Default is an empty list.
	name : str, optional
		Name of the model.
		Default is "unknow_PCTMC"
	
	Returns
	-------
	PCTMC
		the PCTMC describes by `transitions`, `labelling`, and `initial_state`.
	"""
	if 'init' in labelling:
		msg =  "The label 'init' cannot be used: it is reserved for initial states."
		raise SyntaxError(msg)
	
	labelling.append('init')
	
	states = [i[0] for i in transitions]+[i[1] for i in transitions]
	states += [i[0] for i in synchronous_transitions]+[i[2] for i in synchronous_transitions]
	states = list(set(states))
	states.sort()
	nb_states = len(states)
	if type(initial_state) == int:
		transitions.append((nb_states,initial_state,1.0))
	else:
		for i,j in enumerate(initial_state):
			transitions.append((nb_states,i,j))
	nb_states += 1
	if nb_states > len(labelling):
		raise ValueError("All states are not labelled (the labelling list is too small).")
	elif nb_states < len(labelling):
		print("WARNING: the labelling list is bigger than the number of states")
	
	parameter_str = parameters
	parameter_values = {}
	parameter_indexes = [[] for _ in parameter_str]
	transition_expr = [sympify('0.0')]

	matrix = zeros((nb_states,nb_states),dtype='uint16')
	for t in transitions:
		val = str(t[2])
		expr = sympify(val)
		if expr.is_real or not expr in transition_expr:
			transition_expr.append(expr)
		matrix[t[0],t[1]] = transition_expr.index(expr)
		for p in expr.free_symbols:
			parameter_indexes[parameter_str.index(str(p))].append([t[0],t[1]])

	for p in parameter_instantiation:
		if not p in parameter_str:
			print("WARNING: no parameter "+p+", instantiation ignored.")
		else:
			parameter_values[p] = parameter_instantiation[p]
	return PCTMC(matrix,labelling,transition_expr,parameter_values,
				parameter_indexes,parameter_str,name,synchronous_transitions)

def synchronousCompositionPCTMCs(ms: list, name: str = "unknown_composition") -> PCTMC:
	"""
	Returns the synchornous compotision of the PCTMCs in `ms`.

	Parameters
	----------
	ms : list of PCTMCs
		List of PCTMCs to compose.
	name : str, optional.
		Name of the output model.
		Default is unknown_composition.

	Returns
	-------
	PCTMC
		Synchronous composition of `m1` and `m2`.
	"""
	def removeUnreachableState(m):
		i = 0
		while i < len(m.matrix): # removing unreachable state
			if (m.matrix.T[i] == 0).all() == True and m.labelling[i] != 'init':
				m.nb_states -= 1
				m.matrix = delete(m.matrix,i,0)
				m.matrix = delete(m.matrix,i,1)
				m.initial_state = delete(m.initial_state,i)
				m.labelling = m.labelling[:i]+m.labelling[i+1:]
				for j in range(len(m.parameter_indexes)):
					k = 0
					while k < len(m.parameter_indexes[j]): 
						if i in m.parameter_indexes[j][k]:
							m.parameter_indexes[j] = m.parameter_indexes[j][:k]+m.parameter_indexes[j][k+1:]
						else:
							if m.parameter_indexes[j][k][0] > i:
								m.parameter_indexes[j][k][0] -= 1
							if m.parameter_indexes[j][k][1] > i:
								m.parameter_indexes[j][k][1] -= 1
							k += 1
				i = -1
			i += 1

	def removeUnusedTrans(m):
		to_remove = list(set(range(len(m.transition_expr))) - set(m.matrix.flatten()) )
		for i in to_remove:
			m.transition_expr = m.transition_expr[:i]+m.transition_expr[i+1:]
			for j in range(len(m.matrix)):
				for k in range(len(m.matrix)):
					if m.matrix[j][k] > i:
						m.matrix[j][k] -= 1
			for j in range(len(to_remove)):
				if to_remove[j] > i:
					to_remove[j] -= 1

	if len(ms) == 0:
		raise ValueError("ms must contain at least 2 models.")
	elif len(ms) == 1:
		return ms[0]
	m1 = ms[0]
	sync_trans = []
	for i in range(1,len(ms)):
		for t in sync_trans:
			m1.synchronous_transitions.append(t)
		m1, sync_trans = synchronousComposition2PCTMCs(m1, ms[i], name)
	# add sync trans
	for t in sync_trans:
		i, _, d, expr = t
		if m1.matrix[i,d] != 0.0:
			expr = expr + m1.transitionExpression(i,d)
		if not expr in m1.transition_expr:
			m1.transition_expr.append(expr)
		m1.matrix[i,d] = m1.transition_expr.index(expr)
		for p in expr.free_symbols:
			m1.parameter_indexes[m1.parameter_str.index(p)].append([i,d])
	removeUnreachableState(m1)
	removeUnusedTrans(m1)
	return m1

def synchronousComposition2PCTMCs(m1: PCTMC, m2: PCTMC, name: str = "unknown_composition") -> PCTMC:
	"""
	Returns the synchronous composition of `m1` and `m2`.

	Parameters
	----------
	m1 : CTMC
		First CTMC to compose with.
	m2 : CTMC
		Second CTMC to compose with.

	Returns
	-------
	PCTMC
		Synchronous composition of `m1` and `m2`.
	"""
	m1_init = [i for i,li in enumerate(m1.labelling) if li == 'init']
	m2_init = [i for i,li in enumerate(m2.labelling) if li == 'init']
	m1_nb_states = m1.nb_states - len(m1_init)
	m2_nb_states = m2.nb_states - len(m2_init)
	nb_states = m1_nb_states * m2_nb_states	
	m1_sids = [i-m1.labelling[:i].count("init") for i,li in enumerate(m1.labelling) if li != 'init']
	m2_sids = [i-m1.labelling[:i].count("init") for i,li in enumerate(m2.labelling) if li != 'init']

	matrix= zeros((nb_states,nb_states),dtype='uint16')
	labelling = []
	p_v = m1.parameter_values #TODO m2.parameter_values
	p_str = list(set(m1.parameter_str+m2.parameter_str))
	p_i = [[] for _ in p_str]
	trans_expr = [sympify('0.0')]

	def get_state_index(s1,s2):
		s2 = m2_sids.index(s2)
		s1 = m1_sids.index(s1)
		return s1*m2_nb_states+s2

	def add_in_matrix(trans,s1,s2,model,add_index=[]):
		if model == 1:
			for i in m2_sids:
				x,y = get_state_index(s1,i),get_state_index(s2,i)
				matrix[x,y] = trans
				for ind in add_index:
					p_i[ind].append([x,y])
		elif model == 2:
			for i in m1_sids:
				x,y = get_state_index(i,s1),get_state_index(i,s2)
				matrix[x,y] = trans
				for ind in add_index:
					p_i[ind].append([x,y])
	
	def add_transition(i,j,model):
		if model == 1:
			m = m1
		else:
			m = m2
		expr = m.transitionExpression(i,j)
		if not expr in trans_expr:
			trans_expr.append(expr)
		add_index = []
		for p in m.involvedParameters(i,j):
			if not p in p_str:
				p_str.append(p)
			add_index.append(p_str.index(p))
		add_in_matrix(trans_expr.index(expr),i,j,model,add_index)
	
	def add_sync_transition(a,s1, d1, expr1, s2, d2, expr2):
		expr1 = sympify(expr1)
		expr2 = sympify(expr2)
		expr = expr1*expr2
		for p in expr.free_symbols:
			if not p in p_str:
				p_str.append(p)
				p_i.append([])
		x, y = get_state_index(s1,s2), get_state_index(d1, d2)
		return [x,a,y,expr]

	for s1 in m1_sids: # labelling
		for s2 in m2_sids:
			labelling.append(m1.labelling[s1]+','+m2.labelling[s2])
	for i in m1_sids: # m1 transitions
		for j in m1_sids:
			add_transition(i,j,1)
	for i in m2_sids: # m2 transitions
		for j in m2_sids:
			add_transition(i,j,2)
	for i in m1_sids: # self loops
		if m1.matrix[i,i] != 0:
			for j in m2_sids:
				if m2.matrix[j,j] != 0:
					expr = m1.transitionExpression(i,i)+m2.transitionExpression(j,j)
					if not expr in trans_expr:
						trans_expr.append(expr)
					matrix[get_state_index(i,j),get_state_index(i,j)] = trans_expr.index(expr)
	sync_trans = []
	for sync_1 in m1.synchronous_transitions: # synchronous transitions
		si,ai,di,pi = sync_1
		for sync_2 in m2.synchronous_transitions:
			sj,aj,dj,pj = sync_2
			if ai == aj:
				sync_trans.append(add_sync_transition(ai,si,di,pi,sj,dj,pj))
				
	labelling.append('init')
	matrix = vstack((matrix,zeros(nb_states,dtype='uint16')))
	matrix = hstack((matrix,zeros(nb_states+1,dtype='uint16')[:,newaxis]))
	m1_init_trans = zeros(m1_nb_states)
	for i in m1_init:
		tmp = [m1.transitionValue(i,j) for j in m1_sids]
		for j in range(m1_nb_states):
			if m1_init_trans[j]*tmp[j]>0.0:
				m1_init_trans[j] *= tmp[j]
			else:
				m1_init_trans[j] += tmp[j]
	m2_init_trans = zeros(m2_nb_states)
	for i in m2_init:
		tmp = [m2.transitionValue(i,j) for j in m2_sids]
		for j in range(m2_nb_states):
			if m2_init_trans[j]*tmp[j]>0.0:
				m2_init_trans[j] *= tmp[j]
			else:
				m2_init_trans[j] += tmp[j]
	for i,si in enumerate(m1_init_trans):
		for j,sj in enumerate(m2_init_trans):
			if si*sj == 0.0:
				matrix[-1][get_state_index(i,j)] = 0
			else:
				matrix[-1][get_state_index(i,j)] = len(trans_expr)
				trans_expr.append(sympify(str(si*sj)))

	return PCTMC(matrix,labelling,trans_expr,p_v,p_i,p_str,name), sync_trans



import unittest
from ..gohmm import *
from os import remove
from ..base.Set import *
from ..base.BW import BW
from math import exp, sqrt, pi

def modelGoHMM():
	initial_state = [0.1,0.7,0.0,0.0,0.2]
	transitions = [(0,0,0.9),(0,1,0.1),(1,0,0.05),(1,1,0.9),(1,2,0.04),
				   (1,4,0.01),(2,1,0.05),(2,2,0.8),(2,3,0.14),(2,4,0.01),
				   (3,2,0.05),(3,3,0.95),(4,1,0.1),(4,4,0.9)]
	output = [(0,3.0,5.0),(0,2.0,4.0),(1,0.5,1.5),(1,2.5,1.5),(2,0.2,0.7),
			  (2,1.0,1.0),(3,0.0,0.3),(3,1.5,5.0),(4,2.0,4.0),(4,0.5,0.5)]
	return createGoHMM(transitions,output,initial_state,name="My_GoHMM")

m = modelGoHMM()

class GoHMMTestclass(unittest.TestCase):

	def test_GoHMM_state(var):
		mu = m.mu(0)
		b = exp(-0.02)/(5.0*sqrt(2*pi))*exp(-1/32)/(4.0*sqrt(2*pi))
		sigma = [m.output[0][i][1] for i in range(2)]
		
		var.assertEqual(m.a(0,0),0.9)
		var.assertEqual(m.a(0,2),0.0)
		var.assertEqual(sigma,[5.0,4.0])
		var.assertTrue(all(mu==array([3.0,2.0])))
		var.assertEqual(m.b(0,[4.0,3.0]),b)
		var.assertEqual(m.tau(0,0,[4.0,3.0]),0.9*b)
	
	def test_GoHMM_save_load_str(var):
		m.save("test_save.txt")
		mprime = loadGoHMM("test_save.txt")
		var.assertEqual(type(m),type(mprime))
		var.assertEqual(type(m),GoHMM)
		
		var.assertEqual(str(m),str(mprime))
		remove("test_save.txt")
	
	def test_GoHMM_Set(var):
		set1 = m.generateSet(50,10)
		var.assertEqual(set1.type,3)
		set2 = m.generateSet(50,1/4,"geo",6)
		set1.addSet(set2)
		set1.save("test_save.txt")
		set2 = loadSet("test_save.txt")
		var.assertEqual(set2.type,3)
		var.assertTrue(set1.isEqual(set2))
		remove("test_save.txt")
	
	def test_BW_GoHMM(var):
		initial_model   = loadGoHMM("jajapy/tests/materials/gohmm/random_GoHMM.txt")
		training_set    = loadSet("jajapy/tests/materials/gohmm/training_set_GoHMM.txt")
		output_expected = loadGoHMM("jajapy/tests/materials/gohmm/output_GoHMM.txt")
		output_gotten   = BW().fit( training_set, initial_model)
		test_set = m.generateSet(10000,10)
		var.assertAlmostEqual(output_expected.logLikelihood(test_set),
							  output_gotten.logLikelihood(test_set))

if __name__ == "__main__":
	unittest.main()


import unittest
from ..mdp import *
from os import remove
from ..base.Set import *
from ..base.BW import BW
from math import log
from ..mdp.Active_BW_MDP import Active_BW_MDP
from numpy import where

def modelMDP_bigstreet(p=0.75):
	labelling = ['R','L','R','L','OK','HIT']
	transitions = [(0,'m',1,p),(0,'m',2,1-p),(0,'s',3,p),(0,'s',0,1-p),
				   (1,'m',0,p),(1,'m',3,1-p),(1,'s',2,p),(1,'s',1,1-p),
				   (2,'m',5,1.0),(2,'s',4,1.0),(3,'m',5,1.0),(3,'s',4,1.0),
				   (4,'m',4,1.0),(4,'s',4,1.0),(5,'m',5,1.0),(5,'s',5,1.0)]
	return createMDP(transitions,labelling,0,"bigstreet")

m = modelMDP_bigstreet()
scheduler = UniformScheduler(m.getActions())

class MDPTestclass(unittest.TestCase):
	
	def test_MDP_initial_state(var):
		p = 0.75
		labelling = ['R','L','R','L','OK','HIT']
		transitions = [(0,'m',1,p),(0,'m',2,1-p),(0,'s',3,p),(0,'s',0,1-p),
				   (1,'m',0,p),(1,'m',3,1-p),(1,'s',2,p),(1,'s',1,1-p),
				   (2,'m',5,1.0),(2,'s',4,1.0),(3,'m',5,1.0),(3,'s',4,1.0),
				   (4,'m',4,1.0),(4,'s',4,1.0),(5,'m',5,1.0),(5,'s',5,1.0)]
		mdp = createMDP(transitions,labelling,0)
		var.assertEqual(mdp.nb_states,7)
		var.assertEqual(mdp.labelling.count('init'),1)
		var.assertEqual(mdp.getLabel(int(where(mdp.initial_state == 1.0)[0][0])),'init')
		
		labelling = ['R','L','R','L','OK','HIT']
		mdp = createMDP(transitions,labelling,[0.3,0.0,0.0,0.2,0.5,0.0])
		var.assertEqual(mdp.nb_states,7)
		var.assertEqual(mdp.labelling.count('init'),1)
		var.assertEqual(mdp.pi(6),1.0)
		var.assertTrue((mdp.matrix[-1][0]==array([0.3,0.0,0.0,0.2,0.5,0.0,0.0])).all())
		
		labelling = ['R','L','R','L','OK','HIT']
		mdp = createMDP(transitions,labelling,array([0.3,0.0,0.0,0.2,0.5,0.0]))
		var.assertEqual(mdp.nb_states,7)
		var.assertEqual(mdp.labelling.count('init'),1)
		var.assertEqual(mdp.pi(6),1.0)
		var.assertTrue((mdp.matrix[-1][0]==array([0.3,0.0,0.0,0.2,0.5,0.0,0.0])).all())
		

	def test_MDP_state(var):
		var.assertEqual(m.tau(0,'m',1,'B'),0.0)
		var.assertEqual(m.tau(0,'m',1,'R'),0.75)
		var.assertEqual(m.tau(0,'something else',1,'R'),0.0)
		var.assertEqual(m.tau(0,'m',1,'something else'),0.0)
		var.assertEqual(set(m.getLabel(0)),
						set('R'))
		var.assertEqual(set(m.getActions(0)),
						set(['m','s']))
	
	def test_MDP_save_load_str(var):
		m.save("test_save.txt")
		mprime = loadMDP("test_save.txt")
		var.assertEqual(str(m),str(mprime))
		remove("test_save.txt")
	
	def test_MDP_observations_actions(var):
		var.assertEqual(set(m.getAlphabet()),
						set(['L','R','HIT','OK','init']))
		var.assertEqual(set(m.getActions()),
						set(['m','s']))
	
	def test_MDP_Set(var):
		set1 = m.generateSet(50,10,scheduler)
		set2 = m.generateSet(50,1/4,scheduler,"geo",6)
		set1.addSet(set2)
		var.assertEqual(set1.type,1)
		set1.save("test_save.txt")
		set2 = loadSet("test_save.txt")
		var.assertTrue(set1.isEqual(set2))
		remove("test_save.txt")

	def test_MDP_logLikelihood(var):
		set1 = Set([['init','m','R','m','L','s','L','s','R','m','HIT','m','HIT','s','HIT']],[1],t=1)
		set2 = Set([['init','m','R','s','R','m','R','s','OK','m','OK','m','OK']],[2],t=1)
		l12 = m._logLikelihood_multiproc(set1)
		l11 = m._logLikelihood_oneproc(set1)
		var.assertAlmostEqual(l11,l12)
		var.assertAlmostEqual(l11,log(9/64))
		l2 = m.logLikelihood(set2)
		var.assertAlmostEqual(l2,log(1/16))
		set1.addSet(set2)
		l3 = m.logLikelihood(set1)
		var.assertAlmostEqual(l3,(log(9/64)+2*log(1/16))/3)
	
	def test_BW_MDP(var):
		initial_model   = loadMDP("jajapy/tests/materials/mdp/random_MDP.txt")
		training_set    = loadSet("jajapy/tests/materials/mdp/training_set_MDP.txt")
		output_expected = loadMDP("jajapy/tests/materials/mdp/output_MDP.txt")
		output_gotten   = BW().fit( training_set, initial_model, stormpy_output=False)
		test_set = m.generateSet(10000,10,scheduler)
		var.assertAlmostEqual(output_expected.logLikelihood(test_set),
							  output_gotten.logLikelihood(test_set))
	
	def test_Active_BW_MDP(var):
		initial_model   = loadMDP("jajapy/tests/materials/mdp/random_MDP.txt")
		training_set    = loadSet("jajapy/tests/materials/mdp/training_set_MDP.txt")
		output_expected = loadMDP("jajapy/tests/materials/mdp/active_output_MDP.txt")
		output_gotten   = Active_BW_MDP().fit(training_set, sul=m,initial_model=initial_model, lr=0,
											  nb_iterations=10,nb_sequences=10,sequence_length=10,
											  stormpy_output=False)
		test_set = m.generateSet(10000,10,scheduler)
		var.assertAlmostEqual(output_expected.logLikelihood(test_set),
							  output_gotten.logLikelihood(test_set),places=2)


	
	def test_IOAlergia(var):
		training_set    = loadSet("jajapy/tests/materials/mdp/training_set_MDP.txt")
		IOAlergia().fit(training_set,0.01, stormpy_output=False)

	def test_UniformScheduler(var):
		nb_trials = 100000
		actions = m.getActions()
		results = [0.0 for i in actions]
		for i in range(nb_trials):
			results[actions.index(scheduler.getAction())] += 1/nb_trials
		for i in range(len(results)):
			var.assertAlmostEqual(results[i],len(results)/nb_trials,delta=3)


if __name__ == "__main__":
	unittest.main()


import unittest
from ..pctmc import *
from os import remove
from ..base.Set import *
from ..base.BW import BW
from ..with_stormpy.model_converter import loadPrism

def modelPCTMC_ped():
	labelling = ['red','green']
	transitions = [(1,0,5.0)]
	sync_trans = [(0,'b',1,0.5),(0,'r',0,10.0),(0,'g',0,10.0)]
	parameters = ['r','g']
	return createPCTMC(transitions,labelling,parameters,0,{},sync_trans)

def modelPCTMC_car2():
	labelling = ['red','green']
	transitions = []
	sync_trans = [(0,'g',1,'g'),(1,'r',0,'r'),(0,'b',0,10.0),(1,'b',0,2.0)]
	parameters = ['r','g']
	return createPCTMC(transitions,labelling,parameters,1,{},sync_trans)

def modelPCTMC_car1():
	labelling = ['red','green']
	transitions = []
	sync_trans = [(0,'r',1,'r'),(1,'g',0,'g'),(0,'b',0,10.0),(1,'b',0,2.0)]
	parameters = ['r','g']
	return createPCTMC(transitions,labelling,parameters,0,{},sync_trans)

def modelPCTMCCar_tl():
	labelling = ["c_red","c_orange","c_green"]
	transitions = [(1,0,2.0)]
	sync_trans = [(0,'nothing',2,0.5),(2,'button',1,'p')]
	parameters = ['p']
	return createPCTMC(transitions, labelling, parameters,0,{},sync_trans)

def modelPCTMCPed_tl():
	labelling = ["p_red","p_green"]
	transitions = [(1,0,0.1)]
	sync_trans = [(0,'nothing',0,10.0),(0,'button',1,'p/2')]
	parameters = ['p']
	return createPCTMC(transitions, labelling, parameters,0,{},sync_trans)


m = loadPrism('jajapy/tests/materials/pctmc/tl.pm')

class PCTMCTestclass(unittest.TestCase):
	
	def test_PCTMC_instantiation(var):
		var.assertFalse(m.isInstantiated())
		for i in range(6):
			if i != 1:
				var.assertTrue(m.isInstantiated(i))
			else:
				var.assertFalse(m.isInstantiated(i))
		var.assertEqual(['p'],m.involvedParameters(1))
		var.assertEqual(['p'],m.involvedParameters(1,2))
		m.instantiate(['p'],[1.0])
		var.assertTrue(m.isInstantiated())
		var.assertTrue(m.isInstantiated(1))
		var.assertEqual(m.transitionValue(1,2),1/2)
		var.assertEqual(m.parameterValue('p'),1.0)
	
	def test_PCTMC_state(var):
		var.assertEqual(m.e(0),5.0)
		var.assertEqual(m.e(2),2.1)
		var.assertEqual(m.expected_time(0),0.2)
		var.assertEqual(m.l(0,2,'red'),0.0)
		var.assertEqual(m.l(0,1,'c_red_p_red'),5.0)
		var.assertEqual(m.tau(0,1,'c_red_p_red'),1.0)
		var.assertEqual(m.tau(0,2,'yellow'),0.0)
		var.assertEqual(m.tau(2,3,'c_orange_p_green'),2.0/2.1)
		var.assertEqual(m.tau(2,4,'c_orange_p_green'),0.1/2.1)
		t = 1.0
		lkl = 2.1*exp(-2.1*t)
		var.assertEqual(m.lkl(2,t),lkl)
		var.assertEqual(m.lkl(0,-t),0.0)
	
	def test_PCTMC_getLabel_getAlphabet(var):
		var.assertEqual(set(m.getAlphabet()),
						set(['c_red_p_red','c_green_p_red','c_orange_p_green','c_red_p_green','c_orange_p_red','init']))
		var.assertEqual(m.getLabel(2),'c_orange_p_green')
	
	def test_PCTMC_save_load_str(var):
		m.save("test_save.txt")
		mprime = loadPCTMC("test_save.txt")
		var.assertEqual(str(m),str(mprime))
		remove("test_save.txt")
	
	def test_PCTMC_Set(var):
		m.randomInstantiation()
		set1 = m.generateSet(50,10,timed=True)
		var.assertEqual(set1.type,4)
		set2 = m.generateSet(50,1/4,"geo",6, timed=True)
		set1.addSet(set2)
		var.assertEqual(set1.type,4)
		set1.save("test_save.txt")
		set2 = loadSet("test_save.txt")
		var.assertEqual(set2.type,4)
		var.assertTrue(set1.isEqual(set2))
		remove("test_save.txt")
	
	def test_PCTMC_logLikelihood(var):
		#m = loadPrism('jajapy/tests/materials/pctmc/tl.pm')
		m.instantiate(['p'],[1.0])
		set1 = Set([['init', 'c_red_p_red', 'c_green_p_red', 'c_orange_p_green', 'c_red_p_green', 'c_red_p_red']],[1])
		l11 = m._logLikelihood_oneproc(set1)
		l12 = m._logLikelihood_multiproc(set1)
		var.assertAlmostEqual(l11,l12)
		var.assertAlmostEqual(l11,-0.048790164169432056)
		set2 = Set([['init', 0.51245612, 'c_red_p_red',
					0.20149229043430208, 'c_green_p_red',
					0.8691033110137091, 'c_orange_p_green',
					0.40391041817651124, 'c_red_p_green', 
					17.822666563218853, 'c_red_p_red',],
					['init', 0.497854, 'c_red_p_red',
					0.237901319423925, 'c_green_p_red',
					6.906613946631418, 'c_orange_p_green',
					0.6742555535289178, 'c_red_p_green',
					5.788386627774087, 'c_red_p_red']],[1,1])
		l21 = m.logLikelihood(set2)
		var.assertAlmostEqual(l21,-6.553342509457142)
	
	def test_PCTMC_parameters_learning(var):
		initial_model = loadPrism('jajapy/tests/materials/pctmc/tandem_3.sm')
		initial_model.instantiate(["mu1a","mu1b","mu2","kappa"],[0.5,0.5,0.5,0.5])
		training_set    = loadSet("jajapy/tests/materials/pctmc/training_set_PCTMC.txt")
		output_gotten   = BW().fit_parameters(training_set, initial_model,["mu1a","mu1b","mu2","kappa"])
		output_expected = {'mu1a': 0.19154272606453604, 'mu1b': 1.7589871248425895, 'mu2': 2.287450015307295, 'kappa': 4.0504485338277245}
		var.assertDictEqual(output_expected,output_gotten)


if __name__ == "__main__":
	unittest.main()


from .base import *
from .hmm import *
from .mc import *
from .mdp import *
from .ctmc import *
from .gohmm import *
from .pctmc import *
try:
	from .with_stormpy import *
except ModuleNotFoundError:
	print("WARNING: Stormpy not found.")


import pathlib
from setuptools import setup, find_packages
from codecs import open

HERE = pathlib.Path(__file__).parent

# The text of the README file
with open(HERE / "README.md", encoding='utf-8') as f:
	long_description = f.read()
setup(
    name='jajapy',
    long_description=long_description,
    packages=find_packages(),
    install_requires=['numpy', 'scipy', 'alive-progress','sympy'],
    long_description_content_type="text/markdown",
    version='0.10.4',
    url="",
    description='Baum-Welch for all kind of Markov model',
    author='Raphaël Reynouard',
    author_email="raphal20@ru.is",
    license='MIT',
	classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent"
    ],
    include_package_data=True
)


from ..base.tools import resolveRandom, randomProbabilities, checkProbabilities
from math import log
from ..base.Base_MC import *
from ..base.Set import Set
from .Scheduler import Scheduler
from numpy.random import geometric
from numpy import array, append, dot, zeros, vsplit, ndarray, where, reshape, vstack, append, concatenate
from ast import literal_eval
from multiprocessing import cpu_count, Pool
from random import choices

class MDP(Base_MC):
	"""
	Class representing a MDP.
	"""
	def __init__(self,matrix: ndarray, labelling: list, actions:list, name: str="unknown_MDP"):
		"""
		Create a MDP.

		Parameters
		----------
		matrix : ndarray
			Represents the transition matrix.
			`matrix[s1][act_ID][s2][obs_ID]` is the probability of moving 
			from `s1` to `s2` by executing action of ID `act_ID` and seeing 
			the observation of ID `obs_ID`.
		labelling: list of str
			A list of N observations (with N the nb of states).
			If `labelling[s] == o` then state of ID `s` is labelled by `o`.
			Each state has exactly one label.
		actions: list of str
			The list of all possible actions, such that:
			`actions.index("act")` is the ID of `act`.
		name : str, optional
			Name of the model. Default is "unknow_MDP"
		"""
		self.model_type = MDP_ID
		self.actions = actions
		self.nb_actions = len(self.actions)
		super().__init__(matrix,labelling,name)
		for i in range(self.nb_states):
			for a in range(self.nb_actions):
				if not checkProbabilities(matrix[i][a]):
					msg = "The probability to take a transition from state "
					msg+= str(i)+" executing action "+self.actions[a]+" should be 1.0 or 0.0, here it's "+str(matrix[i][a].sum())
					raise ValueError(msg)

	def getActions(self, state:int =-1) -> list:
		"""
		If state is set, returns the list of all the actions available
		in `state`. Otherwise it returns the actions of the model. 


		Parameters
		----------
		state : int, optional
			a state ID

		Returns
		-------
		list of str
			list of actions
		
		Example
		-------
		>>> model.getActions()
		['A','B','C','D']
		>>> model.getActions(1)
		['A','C']
		"""
		if state == -1:
			return self.actions
		else:
			return list(set(self.actions[i] for i in where(self.matrix[state].sum(axis=1) > 0.0)[0]))


	def tau(self,s1: int,action: str,s2: int,obs: str) -> float:
		"""
		Returns the probability of moving from state ``s1`` executing `action`
		to ``s2`` generating observation ``obs``.

		Parameters
		----------
		s1: int
			source state ID.
		action: str
			An action.
		s2: int
			destination state ID.
		obs: str
			generated observation.
		
		Returns
		-------
		float
			A probability.
		
		Example
		-------
		>>> model.tau(0,'A',1,'a')
		0.6
		>>> model.getLabel(0)
		'a'
		>>> model.tau(0,'A',1,'b')
		0.0
		>>> model.tau(0,'B',1,'b')
		0.0
		>>> model.getActions(0)
		['A']		
		"""
		self._checkStateIndex(s1)
		self._checkStateIndex(s2)
		if obs != self.labelling[s1]:
			return 0.0
		if action not in self.actions:
			return 0.0
		return self.matrix[s1][self.actions.index(action)][s2]
	
	def a(self,s1: int,s2: int, action: str) -> float:
		"""
		Returns the probability of moving from state `s1` to state `s2`,
		just after executing `action`.
		Parameters
		----------
		s1 : int
			ID of the source state.		
		s2 : int
			ID of the destination state.
		action: str
			an action.
		
		Returns
		-------
		float
			Probability of moving from state `s1` to state `s2`.
		
		Example
		-------
		>>> model.a(0,1)
		0.6
		"""
		self._checkStateIndex(s1)
		self._checkStateIndex(s2)
		return self.matrix[s1][self.actions.index(action)][s2]
	
	def next(self,state: int, action: str) -> tuple:
		"""
		Return a state-observation pair according to the distributions 
		described by matrix.

		Parameters
		----------
		state: int
			source state ID.
		action: str
			An action.

		Returns
		-------
		output : (int, str)
			A state-observation pair.

		Example
		-------
		>>> model.next(0,'A')
		(1,'a')
		>>> model.getLabel(0)
		'a'
		>>> model.next(0)
		(1,'a')
		>>> model.next(0)
		(2,'a')
		>>> model.a(0,1,'A')
		0.6
		>>> model.a(0,2,'A')
		0.4
		>>> model.next(0,'C')
		(2,'a')
		>>> model.a(0,2,'C')
		1.0
		"""
		c = resolveRandom(self.matrix[state][self.actions.index(action)])
		return (c, self.labelling[state])
			
	def run(self,number_steps: int,scheduler: Scheduler, current: int = -1) -> list:
		"""
		Simulates a run of length ``number_steps`` of the model under
		``scheduler`` and returns the sequence of actions-observations generated.
		
		Parameters
		----------
		number_steps: int
			length of the simulation.
		current : int, optional.
			If current it set, it starts from the state `current`.
			Otherwise it starts from an initial state.

		Returns
		-------
		output: list of str
			List of alterning state-observation.
		"""
		res = []
		if current == -1:
			current = resolveRandom(self.initial_state)
		scheduler.reset()
		current_len = 0
		while current_len < number_steps:
			action = scheduler.getAction()

			while action not in self.getActions(current):
				action = scheduler.getAction()
			
			next_state, observation = self.next(current,action)
			res.append(observation)
			res.append(action)
			scheduler.addObservation(observation)
			current = next_state
			current_len += 1
		res.append(self.labelling[current])
		return res

	def generateSet(self, set_size: int, param, scheduler: Scheduler, distribution=None, min_size=None) -> list:
		"""
		Generates a set (training set / test set) containing `set_size` traces
		generated under ``scheduler``.

		Parameters
		----------
		set_size: int
			number of traces in the output set.
		param: a list, an int or a float.
			the parameter(s) for the distribution. See "distribution".
		scheduler: Scheduler:
			A scheduler used to generated all the traces.
		distribution: str, optional
			If ``distribution=='geo'`` then the sequence length will be
			distributed by a geometric law such that the expected length is
			``min_size+(1/param)``.
			If distribution==None param can be an int, in this case all the
			seq will have the same length (``param``), or ``param`` can be a
			list of int.
			Default is None.
		min_size: int, optional
			see "distribution". Default is None.
		
		Returns
		-------
		output: list
			a set (training set / test set).
		"""
		seq = []
		val = []
		for i in range(set_size):
			if distribution == 'geo':
				curr_size = min_size + int(geometric(param))
			else:
				if type(param) == list:
					curr_size = param[i]
				elif type(param) == int:
					curr_size = param

			trace = self.run(curr_size, scheduler)

			if not trace in seq:
				seq.append(trace)
				val.append(0)

			val[seq.index(trace)] += 1

		return Set(seq,val,t=1)
	
	def save(self,file_path:str) -> None:
		"""Save the model into a text file.

		Parameters
		----------
		file_path : str
			path of the output file.
		
		Examples
		--------
		>>> model.save("my_model.txt")
		"""
		f = open(file_path, 'w')
		f.write("MDP\n")
		f.write(str(self.actions))
		f.write('\n')
		super()._save(f)
	
	def _stateToString(self,state:int) -> str:
		res = "----STATE "+str(state)+"--"+self.labelling[state]+"----\n"
		for ai,a in enumerate(self.actions):
			for s in range(self.nb_states):
				if self.matrix[state][ai][s] > 0.0001:
					res += "s"+str(state)+" - ("+a+") -> s"+str(s)+" : "+str(self.matrix[state][ai][s])+'\n'
		return res
	
	def savePrism(self,file_path:str) -> None:
		"""
		Save this model into `file_path` in the Prism format.

		Parameters
		----------
		file_path : str
			Path of the output file.
		"""
		f = open(file_path,'w')
		f.write("mdp\n\n")
		super()._savePrism(f)
	
	def _savePrismMatrix(self,f) -> None:
		for s1 in range(self.nb_states):
			for ai,a in enumerate(self.actions):
				if (self.matrix[s1,ai]!=0).any():
					res = '\t['+a+'] s='+str(s1)+' ->'
					for s2 in where(self.matrix[s1,ai] != 0.0)[0]:
						res += ' '+str(self.matrix[s1,ai,s2])+":(s'="+str(s2)+") +"
					res = res[:-2]+';\n'
					f.write(res)
		
	def _logLikelihood_oneproc(self,sequences: Set) -> float:
		"""
		Compute the average loglikelihood of a set of sequences.

		Parameters
		----------
		sequences: Set
			set of sequences of actions-observations.
		
		Returns
		-------
		output: float
			loglikelihood of ``sequences`` under this model.
		"""
		sequences_sorted = sequences.sequences[:]
		sequences_sorted.sort()
		loglikelihood = 0.0
		alpha_matrix = self._initAlphaMatrix(len(sequences_sorted[0])//2)
		for seq in range(len(sequences_sorted)):
			sequence_actions = [sequences_sorted[seq][i+1] for i in range(0,len(sequences_sorted[seq])-1,2)]
			sequence_obs = [sequences_sorted[seq][i] for i in range(0,len(sequences_sorted[seq])-1,2)]
			sequence = sequences_sorted[seq]
			times = sequences.times[sequences.sequences.index(sequence)]
			common = 0
			if seq > 0:
				while common < min(len(sequences_sorted[seq-1]),len(sequence)):
					if sequences_sorted[seq-1][common] != sequence[common]:
						break
					common += 1
			common = int(common/2)
			alpha_matrix = self._updateAlphaMatrix(sequence_obs,sequence_actions,common,alpha_matrix)
			
			last_arr = alpha_matrix[-1] * (array(self.labelling) == sequence[-1])
			last_arr = last_arr.sum()
			if last_arr > 0.0:
				loglikelihood += log(last_arr) * times

		return loglikelihood/sum(sequences.times)
	
	def _updateAlphaMatrix(self, sequence_obs: list,
						   sequence_actions:list,
						   common: int, alpha_matrix: list) -> array:
		"""
		Update the given alpha values for all the states for a new
		`sequence_obs` of observations. It keeps the alpha values for the
		``common`` first observations of the ``sequence``. The idea is the 
		following: if you have already computed the alpha values for a previous
		sequence and you want to compute the alpha values of a new sequence
		that starts with the same ``common`` observations you don't need to
		compute again the first ``common`` alpha values for each states. If,
		on the other hand, you have still not computed any alpha values you can
		simply set ``common`` to 0 and give an empty ``alpha_matrix`` which has
		the right size. The method ``initAlphaMatrix`` can generate such matrix.

		Parameters
		----------
		sequence_obs: list of str
			a sequence of observations.
		sequence_actions: list of str
			a sequence of actions.
		common: int
			for each state, the first ``common`` alpha values will be keept
			unchanged.
		alpha_matrix: 2-D narray of float
			the ``alpha_matrix`` to update. Can be generated by the method
			``initAlphaMatrix``.

		Returns
		-------
		output: 2-D narray of float
			the alpha matrix containing all the alpha values for all the states
			for this sequence: ``alpha_matrix[s][t]`` is the probability of
			being in state ``s`` after seing the ``t-1`` first observation of
			``sequence``.
		"""
		diff_size = len(alpha_matrix)-1 - len(sequence_obs)
		if diff_size < 0: # alpha_matrix too small
			n = zeros(-diff_size * self.nb_states).reshape(-diff_size,self.nb_states)
			alpha_matrix = append(alpha_matrix,n,axis=0)
		elif diff_size > 0: #alpha_matrix too big
			alpha_matrix = vsplit(alpha_matrix,[len(alpha_matrix)-diff_size,self.nb_states])[0]
		for k in range(common,len(sequence_obs)):
			for s in range(self.nb_states):
				p = array([self.tau(ss,sequence_actions[k],s,sequence_obs[k]) for ss in range(self.nb_states)])
				alpha_matrix[k+1,s] = dot(alpha_matrix[k],p)
		return alpha_matrix
	
	def _logLikelihood_multiproc(self, sequences: Set) -> float:
		p = Pool(processes = cpu_count()-1)
		tasks = []
		for seq,times in zip(sequences.sequences,sequences.times):
			tasks.append(p.apply_async(self._computeAlphas, [seq, times,]))
		temp = [res.get() for res in tasks if res.get() != False]
		return sum(temp)/sum(sequences.times)
	
	def _computeAlphas(self,sequence: list, times: int) -> float:
		"""
		Compute the alpha values for ``sequence``.

		Parameters
		----------
		sequence: list of str
			Sequence of alternating actions-observations.
		times: int
			Number of times this sequence appears in the sample.

		Returns
		-------
		float
			loglikelihood of ``sequence`` multiplied by ``times``.
		"""
		len_seq = len(sequence)
		prev_arr = array(self.initial_state)
		
		for k in range(0,len_seq-1,2):
			new_arr = zeros(self.nb_states)
			for s in range(self.nb_states):
				p = array([self.tau(ss,sequence[k+1],s,sequence[k]) for ss in range(self.nb_states)])
				new_arr[s] = dot(prev_arr,p)
			prev_arr = new_arr
		prev_arr = prev_arr*(array(self.labelling) == sequence[-1])
		if prev_arr.sum() == 0.0:
			return 0.0
		return log(prev_arr.sum())*times

def loadMDP(file_path: str) -> MDP:
	"""
	Load an MDP saved into a text file.

	Parameters
	----------
	file_path : str
		Location of the text file.
	
	Returns
	-------
	output : MDP
		The MDP saved in `file_path`.
	"""
	f = open(file_path,'r')
	l = f.readline()[:-1] 
	if l != "MDP":
		msg = " this file doesn't describe an MC: it describes a "+l
		raise ValueError(msg)
	actions = literal_eval(f.readline()[:-1])
	labelling = literal_eval(f.readline()[:-1])
	name = f.readline()[:-1]
	initial_state = array(literal_eval(f.readline()[:-1]))
	matrix = literal_eval(f.readline()[:-1])
	matrix = array(matrix)
	f.close()
	return MDP(matrix, labelling, actions, name)


def MDP_random(nb_states: int,labelling: list, actions: list,random_initial_state: bool = True, deterministic: bool = False) -> MDP:
	"""
	Generate a random MDP.

	Parameters
	----------
	number_states : int
		Number of states.
	labelling : list of str
		List of observations.
	actions : list of str
		List of actions.	
	random_initial_state: bool, optional
		If set to True we will start in each state with a random probability, otherwise we will always start in state 0.
		Default is True.
	deterministic: bool, optional
		If True, the model will be determinstic: in state `s`, with action `a`, for all label `o`,
		there is at most one transition to a state labelled with `o`.
		Default is False.
	
	Returns
	-------
	MDP
		A pseudo-randomly generated MDP.
	"""
	alphabet = list(set(labelling))
	matrix = []
	for s in range(nb_states):
		if not deterministic:
			p = array([append(randomProbabilities(nb_states),0.0) for a in actions])
			p = reshape(p,(len(actions),nb_states+1))
			matrix.append(p)
		else:
			matrix.append([])
			for a in actions:
				dest = [choices(where(array(labelling) == o)[0]) for o in alphabet]
				p = zeros(nb_states)
				probs = randomProbabilities(len(alphabet))
				for i,j in zip(dest,probs):
					p[i] = j
				matrix[-1].append(append(p,0.0))

	if random_initial_state:
		init = randomProbabilities(nb_states).tolist()+[0.0]
	else:
		init = [1.0]+[0.0 for i in range(nb_states)]
	matrix.append(array([init for a in actions]))
	matrix = array(matrix)
	
	labelling = labelsForRandomModel(nb_states,labelling)
	return MDP(matrix, labelling, actions, "MDP_random_"+str(nb_states)+"_states")


def createMDP(transitions:list, labelling:list, initial_state, name: str ="unknown_MDP") -> MDP:
	"""
	An user-friendly way to create an MDP.

	Parameters
	----------
	transitions : [ list of tuples (int, str, int, float)]
		Each tuple represents a transition as follow: 
		(source state ID, action, destination state ID, probability).
	labelling: list of str
		A list of N observations (with N the nb of states).
		If `labelling[s] == o` then state of ID `s` is labelled by `o`.
		Each state has exactly one label.
	initial_state : int or list of float
		Determine which state is the initial one (then it's the id of the
		state), or what are the probability to start in each state (then it's
		a list of probabilities).
	name : str, optional
		Name of the model.
		Default is "unknow_MC"
	
	Returns
	-------
	MDP
		the MDP describes by `transitions`, `labelling`, and `initial_state`.
	
	Examples
	--------
	"""
	if 'init' in labelling:
		msg =  "The label 'init' cannot be used: it is reserved for initial states."
		raise SyntaxError(msg)

	states = list(set([i[0] for i in transitions]+[i[2] for i in transitions]))
	states.sort()
	nb_states = len(states)
	actions = list(set([i[1] for i in transitions]))
	actions.sort()
	nb_actions = len(actions)
	
	if nb_states > len(labelling):
		raise ValueError("all states are not labelled (the labelling list is too small).")
	elif nb_states < len(labelling):
		print("WARNING: the labelling list is bigger than the number of states")

	res = zeros((nb_states,nb_actions,nb_states))
	for t in transitions:
		res[states.index(t[0])][actions.index(t[1])][states.index(t[2])] = t[3]
	
	labelling.append('init')
	res = vstack((res,zeros((1,nb_actions,nb_states))))
	res = concatenate((res,zeros((nb_states+1,nb_actions,1))),axis=2)
	if type(initial_state) == int:
		for a in range(nb_actions):
			res[-1][a][initial_state] = 1.0
	else:
		if type(initial_state) == ndarray:
			initial_state = initial_state.tolist()
		for a in range(nb_actions):
			res[-1][a] = array(initial_state+[0.0])
	
	return MDP(res,labelling,actions,name)


from numpy.random import normal
from numpy import array, zeros, ndarray
from ast import literal_eval
from ..base.tools import randomProbabilities
from ..base.Base_HMM import Base_HMM, GOHMM_ID
from math import sqrt, exp, pi
from random import uniform
	
class GoHMM(Base_HMM):
	"""
	Creates a GoHMM.

	Parameters
	----------
	matrix : ndarray
			Represents the transition matrix.
			`matrix[s1][s2]` is the probability of moving from `s1` to `s2`.
	output : ndarray
			Contains the parameters of the guassian distributions.
			`output[s1][0][0]` is the mu parameter of the first distribution in `s1`,
			`output[s1][0][1]` is the sigma parameter of the first distribution in `s1`.
			`output[s1][1][0]` is the mu parameter of the second distribution in `s1`.
			etc...
	initial_state : int or list of float
		Determine which state is the initial one (then it's the id of the
		state), or what are the probability to start in each state (then it's
		a list of probabilities).
	name : str, optional
		Name of the model.
		Default is "unknown_GoHMM"
	"""
	def __init__(self,matrix,output,initial_state,name="unknown_GoHMM"):
		self.model_type = GOHMM_ID
		self.nb_distributions = len(output[0])
		for i in range(len(output)):
			if len(output[i]) != self.nb_distributions:
				raise ValueError("All state must have as much distributions")

		self.output = array(output)
		try:
			if min(self.output.T[1].flatten()) < 0.0:
				raise ValueError("The sigma parameters must be positive")
		except IndexError:
			raise ValueError("All distribution must have two parameters")
		
		super().__init__(matrix,initial_state,name)
		if len(self.output.flatten()) != self.nb_distributions*self.nb_states*2:
			raise ValueError("All distribution must have two parameters")
	
	def mu(self,s: int) -> ndarray:
		"""
		Returns the mu parameters for this state.

		Parameters
		----------
		s : int
			ID of the source state.

		Returns
		-------
		ndarray
			the mu parameters.
		"""
		return self.output[s].T[0]
			
	def mu_n(self,s: int, n: int) -> float:
		"""
		Returns the mu parameters of the `n`th distribution in `s`.

		Parameters
		----------
		s : int
			ID of the source state.
		n : int
			Index of the distribution.

		Returns
		-------
		float
			the mu parameter of the `n`th distribution in `s`.
		"""
		return self.output[s][n][0]

	
	def b(self,s: int, l: list) -> float:
		"""
		Returns the likelihood of generating `l` in state `s`.

		Parameters
		----------
		s : int
			ID of the source state.		
		l : list of float
			list of observations.
		
		Returns
		-------
		output : float
			Likelihood of generating `l` in state `s`.
		"""
		res = 1.0
		for n,ln in enumerate(l):
			res *= self.b_n(s,n,ln)
		return res

	def b_n(self,s:int,n:int,l:float) -> float:
		"""
		Returns the likelihood of generating, from the nth distribution of `s`,
		observation `l`.

		Parameters
		----------
		s : int
			state ID
		n : int
			Index of the distribution.
		l : str
			The observation.
			
		Returns
		-------
		output : float
			Likelihood of generating, from the `n`th distribution of this
			state, observation `l`.
		"""
		mu, sigma  = self.output[s][n]
		return exp(-0.5*((mu-l)/sigma)**2)/(sigma*sqrt(2*pi))

	def next_obs(self, s:int) -> list:
		"""
		Generates n observations according to the n normal distributions in 
		`s`.

		Parameters
		----------
		s : int
			ID of the source state.	

		Returns
		-------
		output : str
			An observation.
		"""
		return [normal(parameter[0],parameter[1],1)[0] for parameter in self.output[s]]
	
	def tau(self,s1:int,s2:int,obs: list) -> float:
		"""
		Returns the likelihood of generating, from `s1`, observation `obs` 
		while moving to state `s2`.

		Parameters
		----------
		s1 : int
			A state ID.
		s2 : int
			A state ID.
		obs : list of floats
			An observation.

		Returns
		-------
		output : float
			The likelihood of generating, from from `s1`, observation `obs` 
			while moving to state `s2`.
		"""
		return super().tau(s1,s2,obs)

	def save(self,file_path:str):
		"""Save the model into a text file.

		Parameters
		----------
		file_path : str
			path of the output file.
		
		Examples
		--------
		>>> model.save("my_model.txt")
		"""
		f = open(file_path, 'w')
		f.write("GoHMM\n")
		super()._save(f)

	def _stateToString(self,state:int) -> str:
		res = super()._stateToString(state)
		for n in range(self.nb_distributions):
			res += "mean "+str(n+1)+": "+str(round(self.output[state][n][0],4))+'\n'
			res += "std "+str(n+1)+": "+str(round(self.output[state][n][1],4))+'\n'
		return res


def loadGoHMM(file_path: str) -> GoHMM:
	"""
	Load an GoHMM saved into a text file.

	Parameters
	----------
	file_path : str
		Location of the text file.
	
	Returns
	-------
	output : GoHMM
		The GoHMM saved in `file_path`.
	"""
	f = open(file_path,'r')
	l = f.readline()[:-1] 
	if l != "GoHMM":
		print("ERROR: this file doesn't describe an GoHMM: it describes a "+l)
	output = literal_eval(f.readline()[:-1])
	output = array(output)
	name = f.readline()[:-1]
	initial_state = array(literal_eval(f.readline()[:-1]))
	matrix = literal_eval(f.readline()[:-1])
	matrix = array(matrix)
	f.close()
	return GoHMM(matrix, output, initial_state, name)

def GoHMM_random(nb_states:int,nb_distributions:int,
				  random_initial_state:bool=False,
				  min_mu: float=0.0,max_mu: float=2.0,
				  min_sigma: float=0.5,max_sigma: float=2.0) -> GoHMM:
	"""
	Generates a random GoHMM.

	Parameters
	----------
	nu_states : int
		Number of states.
	nb_distributions : int
		Number of distributions in each state.
	alphabet : list of str
		List of observations.
	random_initial_state: bool, optional
		If set to True we will start in each state with a random 
		probability, otherwise we will always start in state 0.
		Default is False.
	min_mu : float, optional
		lower bound for mu. By default 0.0
	max_mu : float, optional
		upper bound for mu. By default 2.0
	min_sigma : float, optional
		lower bound for sigma. By default 0.5
	max_sigma : float, optional
		upper bound for sigma. By default 2.0

	Returns
	-------
	GoHMM
		A pseudo-randomly generated GoHMM.
	"""
	matrix = []
	output = []
	for s in range(nb_states):
		p1 = array(randomProbabilities(nb_states))
		matrix.append(p1)
		p2 = [[round(uniform(min_mu,max_mu),3),round(uniform(min_sigma,max_sigma),3)] for i in range(nb_distributions)]
		p2 = array(p2)
		output.append(p2)
	matrix = array(matrix)
	output = array(output)

	if random_initial_state:
		init = randomProbabilities(nb_states)
	else:
		init = 0
	return GoHMM(matrix, output, init,"GoHMM_random_"+str(nb_states)+"_states")


def createGoHMM(transitions:list, output:list, initial_state:str,name :str = 'unknown_GoHMM') -> GoHMM:
	"""
	An user-friendly way to create a GoHMM.

	Parameters
	----------
	transitions : [ list of tuples (int, int, float)]
		Each tuple represents a transition as follow: 
		(source state ID, destination state ID, probability).
	output : list of list of tuples (float, float)]
		Represents the parameters of the gaussian distributions 
		`[(mu1, sigma1),(mu2, sigma2),...]`.
		`output[0]` contains the parameters of the distributions in state 0
		`output[0][0]` contains the 2 parameters of the first distribution
		in state 0.
		`output[0][0][0]` is the mu parameter of the first distribution in
		state 0, and `output[0][0][1]` is the sigma parameter of the first
		distribution in state 0.
	initial_state : int or list of float
		Determine which state is the initial one (then it's the id of the
		state), or what are the probability to start in each state (then it's
		a list of probabilities).
	name : str, optional
		Name of the model.
		Default is "unknow_GoHMM".
	
	Returns
	-------
	GoHMM
		the GoHMM describes by `transitions`, `emission`, and `initial_state`.
	
	Examples
	--------
	"""
	states = list(set([i[0] for i in transitions]+[i[1] for i in transitions]))
	states.sort()
	nb_distributions = len(output[0])
	nb_states = len(states)
	matrix = zeros((nb_states,nb_states))
	for t in transitions:
		matrix[states.index(t[0])][states.index(t[1])] = t[2]
	output = array((nb_states,nb_distributions,2))
	for s,ts in enumerate(output):
		for i,t in enumerate(ts):
			output[s,i][0] = t[0]
			output[s,i][1] = t[1]
	return GoHMM(matrix, output,initial_state,name)


from numpy.random import exponential
from ast import literal_eval
from ..base.tools import resolveRandom, randomProbabilities
from ..base.Set import Set
from ..mc.MC import MC
from ..pctmc.PCTMC import PCTMC
from ..base.Base_MC import *
from math import exp, log
from random import randint
from numpy import array, zeros, dot, ndarray, vstack, hstack, newaxis, append, delete, where, insert
from sys import platform
from multiprocessing import cpu_count, Pool
from sympy import sympify

class CTMC(Base_MC):
	"""
	Class representing a CTMC.
	"""
	def __init__(self, matrix: ndarray, labelling: list,
				 name: str ="unknown_CTMC",
				 synchronous_transitions: list =[]) -> None:
		"""
		Creates an CTMC.

		Parameters
		----------
		matrix : ndarray
			A (N x N) ndarray (with N the nb of states).
			Represents the transition matrix.
			`matrix[s1][s2]` is the rate associated to the transition 
			from `s1` to `s2`.
		labelling: list of str
			A list of N observations (with N the nb of states).
			If `labelling[s] == o` then state of ID `s` is labelled by `o`.
			Each state has exactly one label.
		name : str, optional
			Name of the model.
			Default is "unknow_CTMC"
		synchronous_transitions: list, optional.
			This is useful only for synchronously composing this CTMC with
			another one.
			List of (source_state <int>, action <str>, dest_state <int>, rate <float>).
			Default is an empty list.
		"""
		self.model_type = CTMC_ID
		self.synchronous_transitions = synchronous_transitions

		super().__init__(matrix,labelling,name)
		for s in range(self.nb_states):
			synchronous_transitions_source = [i[0] for i in synchronous_transitions]
			if self.e(s) == 0.0 and s not in synchronous_transitions_source:
				print("WARNING: State "+str(s)+" doesn't have any leaving transition.")

	def e(self,s: int) -> float:
		"""
		Returns the exit rate of state ``s``, i.e. the sum of all the rates in
		this state.

		Returns
		-------
		s : int
			A state ID.
		float
			An exit rate.
		"""
		self._checkStateIndex(s)
		return sum(self.matrix[s])

	
	def l(self, s1:int, s2:int, obs:str) -> float:
		"""
		Returns the rate associated to the transition from state `s1`, seeing
		`obs`, to state `s2`.

		Parameters
		----------
		s1 : int
			A state ID.
		s2 : int
			A state ID.
		obs : str
			An observation.

		Returns
		-------
		float
			A rate.
		"""
		self._checkStateIndex(s1)
		self._checkStateIndex(s2)
		if self.labelling[s1] != obs:
			return 0.0
		return self.matrix[s1][s2]
	
	def lkl(self, s: int, t: float) -> float:
		"""
		Returns the likelihood of staying in `s` state for a duration `t`.

		Parameters
		----------
		s : int
			A state ID.
		t : float
			A waiting time.

		Returns
		-------
		float
			A Likelihood.
		"""
		if t < 0.0:
			return 0.0
		return self.e(s)*exp(-self.e(s)*t)
	
	def tau(self, s1:int, s2: int, obs: str) -> float:
		"""
		Returns the probability to move from `s1` to `s2` generating
		observation ``obs``.

		Parameters
		----------
		s1 : int
			A state ID.
		s2 : int
			A state ID.
		obs : str
			An observation.

		Returns
		-------
		float
			A probability.
		"""
		return self.l(s1,s2, obs)/self.e(s1)
	
	def expected_time(self, s:int) -> float:
		"""
		Returns the expected waiting time in `s`, i.e. the inverse of
		the exit rate.
		
		Parameters
		----------
		s : int
			A state ID.
		
		Returns
		-------
		float
			expected waiting time in this state.
		"""
		return 1/self.e(s)

	def _stateToString(self,state:int) -> str:
		res = "----STATE "+str(state)+"--"+self.labelling[state]+"----\n"
		res += "Exepected waiting time: "+str(self.expected_time(state))+'\n'
		den = self.e(state)
		for j in range(len(self.matrix[state])):
			if self.matrix[state][j]/den > 0.0001:
				res += "s"+str(state)+" -> s"
				res += str(j)+" : lambda = "+str(self.matrix[state][j])+'\n'
		return res
	
	def next(self,state: int) -> tuple:
		"""
		Return a state-observation pair according to the distributions 
		described by matrix
		
		Returns
		-------
		output : (int, str)
			A state-observation pair.
		"""
		exps = []
		for exp_lambda in self.matrix[state]:
			if exp_lambda <= 0.0:
				exps.append(1024)
			else:
				exps.append(exponential(1/exp_lambda))
		next_state= exps.index(min(exps))
		next_obs = self.labelling[state]
		return (next_obs, next_state, min(exps))

	def run(self,number_steps: int, timed: bool = False) -> list:
		"""
		Simulates a run of length ``number_steps`` of the model and return the
		sequence of observations generated. If ``timed`` it returns a list of
		pairs waiting time-observation.
		
		Parameters
		----------
		number_steps: int
			length of the simulation.
		timed: bool, optional
			Wether or not it returns also the waiting times. Default is False.

		Returns
		-------
		output: list of str
			trace generated by the run.
		"""
		output = []
		current = resolveRandom(self.initial_state)
		c = 0
		while c < number_steps:
			[symbol, next_state, time_spent] = self.next(current)
			output.append(symbol)
			if timed:
				output.append(time_spent)
			current = next_state
			c += 1
		output.append(self.labelling[current])
		return output
	
	def _computeAlphas_timed(self, sequence: list, times: int) -> float:
		obs_seq   = [sequence[i] for i in range(0,len(sequence),2)]
		times_seq = [sequence[i] for i in range(1,len(sequence),2)]
		len_seq = len(obs_seq)-1
		prev_arr = array(self.initial_state)
		for k in range(len_seq):
			new_arr = zeros(self.nb_states)
			for s in range(self.nb_states):
				p = array([self.l(ss,s,obs_seq[k])*exp(-self.e(ss)*times_seq[k]) for ss in range(self.nb_states) ])
				new_arr[s] = dot(prev_arr,p)
			prev_arr = new_arr
		last_arr = prev_arr * (array(self.labelling) == obs_seq[-1])
		try:
			res = log(last_arr.sum())*times
		except ValueError:
			return False
		return res

	def logLikelihood(self,traces: Set) -> float:
		"""
		Computes the loglikelihood of `traces` under this model.

		Parameters
		----------
		traces : Set
			a set of traces.

		Returns
		-------
		float
			the loglikelihood of `traces` under this model.
		"""
		if traces.type == 0: # non-timed traces
			return super().logLikelihood(traces)
		else: # timed traces
			if platform != "win32" and platform != "darwin":
				p = Pool(processes = cpu_count()-1)
				tasks = []
				for seq,times in zip(traces.sequences,traces.times):
					tasks.append(p.apply_async(self._computeAlphas_timed, [seq, times,]))
				temp = [res.get() for res in tasks if res.get() != False]
			else:
				temp = [self._computeAlphas_timed(seq,times) for seq,times in zip(traces.sequences,traces.times)]
				while False in temp:
					temp.remove(False)
			if len(temp) == 0:
				return -9999
			return sum(temp)/sum(traces.times)

	def toMC(self, name: str="unknown_MC") -> MC:
		"""
		Returns the equivalent untimed MC.

		Parameters
		----------
		name : str, optional
			Name of the output model. By default "unknown_MC"

		Returns
		-------
		MC
			An equivalent untimed model.
		"""
		new_matrix = self.matrix
		for i in range(self.nb_states):
			new_matrix[i] /= self.e(i)

		return MC(new_matrix,self.labelling,name)

	def save(self,file_path:str) -> None:
		"""Save the model into a text file.

		Parameters
		----------
		file_path : str
			path of the output file.
		
		Examples
		--------
		>>> model.save("my_model.txt")
		"""
		f = open(file_path, 'w')
		f.write("CTMC\n")
		super()._save(f)
	
	def savePrism(self,file_path:str) -> None:
		"""
		Save this model into `file_path` in the Prism format.

		Parameters
		----------
		file_path : str
			Path of the output file.
		"""
		f = open(file_path,'w')
		f.write("ctmc\n\n")
		super()._savePrism(f)
	
	def _savePrismMatrix(self,f) -> None:
		for s1 in range(self.nb_states):
			if (self.matrix[s1]!=0).any():
				res = '\t[] s='+str(s1)+' ->'
				for s2 in where(self.matrix[s1] != 0.0)[0]:
					res += ' '+str(self.matrix[s1,s2])+":(s'="+str(s2)+") +"
				res = res[:-2]+';\n'
				f.write(res)

def loadCTMC(file_path: str) -> CTMC:
	"""
	Load an CTMC saved into a text file.

	Parameters
	----------
	file_path : str
		Location of the text file.
	
	Returns
	-------
	output : CTMC
		The CTMC saved in `file_path`.
	"""
	f = open(file_path,'r')
	l = f.readline()[:-1] 
	if l != "CTMC":
		print("ERROR: this file doesn't describe an CTMC: it describes a "+l)
	labelling = literal_eval(f.readline()[:-1])
	name = f.readline()[:-1]
	initial_state = array(literal_eval(f.readline()[:-1]))
	matrix = literal_eval(f.readline()[:-1])
	matrix = array(matrix)
	f.close()
	return CTMC(matrix, labelling, name)

def CTMC_random(nb_states: int, labelling: list, min_exit_rate_time : int,
				max_exit_rate_time: int, self_loop: bool = True,
				random_initial_state: bool=True) -> CTMC:
	"""
	Generates a random CTMC. All the rates will be between 0 and 1.
	All the exit rates will be integers.

	Parameters
	----------
	nb_states : int
		Number of states.
	labelling : list of str
		List of observations.
	min_exit_rate_time: int
		Minimum exit rate for the states (included).
	max_exit_rate_time: int
		Maximum exit rate for the states (included).
	self_loop: bool, optional
		Wether or not there will be self loop in the output model.
		Default is True.
	random_initial_state: bool, optional
		If set to True we will start in each state with a random probability, otherwise we will always start in state 0.
		Default is True.
	
	Returns
	-------
	CTMC
		A pseudo-randomly generated CTMC.

	Examples
	--------
	>>> model = CTMC_random(2,['a','b'],1,5)
	>>> print(model)
	Name: CTMC_random_2_states
	Initial state: s2
	----STATE 0--a----
	Exepected waiting time: 2.0
	s0 -> s0 : lambda = 0.38461538461538464
	s0 -> s1 : lambda = 0.11538461538461539
	----STATE 1--b----
	Exepected waiting time: 4.0
	s1 -> s0 : lambda = 0.13636363636363635
	s1 -> s1 : lambda = 0.11363636363636363
	----STATE 2--init----
	Exepected waiting time: 1.0
	s2 -> s0 : lambda = 0.2
	s2 -> s1 : lambda = 0.8
	"""
	matrix = zeros((nb_states+1,nb_states))
	for i in range(nb_states):
		if self_loop:
			matrix[i] = array(randomProbabilities(nb_states))
		else:
			p = randomProbabilities(nb_states-1).tolist()
			p = insert(p,i,0.0)
			matrix[i] = array(p)
		av_waiting_time = randint(min_exit_rate_time,max_exit_rate_time)
		matrix[i] /= av_waiting_time

	if random_initial_state:
		matrix[-1] = randomProbabilities(nb_states)
	else:
		matrix[-1] = array([1.0]+[0.0 for i in range(nb_states-1)])
	
	matrix = hstack((matrix,zeros(len(matrix))[:,newaxis]))
	
	labelling = labelsForRandomModel(nb_states,labelling)
	return CTMC(matrix, labelling,"CTMC_random_"+str(nb_states)+"_states")

def createCTMC(transitions: list, labelling: list, initial_state,
			   name: str ="unknown_CTMC",synchronous_transitions: list =[]) -> CTMC:
	"""
	An user-friendly way to create a CTMC.

	Parameters
	----------
	transitions : [ list of tuples (int, int, float)]
		Each tuple represents a transition as follow: 
		(source state ID, destination state ID, rate).
	labelling: list of str
		A list of N observations (with N the nb of states).
		If `labelling[s] == o` then state of ID `s` is labelled by `o`.
		Each state has exactly one label.
	initial_state : int or list of float
		Determine which state is the initial one (then it's the id of the
		state), or what are the probability to start in each state (then it's
		a list of probabilities).
	name : str, optional
		Name of the model.
		Default is "unknow_CTMC"
	synchronous_transitions: list, optional.
		This is useful only for synchronously composing this CTMC with
		another one.
		List of (source_state <int>, action <str>, dest_state <int>, rate <float>).
		Default is an empty list.
	
	Returns
	-------
	CTMC
		the CTMC describes by `transitions`, `labelling`, and `initial_state`.
	
	Examples
	--------
	>>> model = createCTMC([(0,1,1.0),(1,0,0.3),(1,1,0.2)],['b','a'],0,"My_MC")
	>>> print(model)
	Name: My_MC
	Initial state: s0
	----STATE 0--b----
	Exepected waiting time: 1.0
	s0 -> s1 : lambda = 1.0
	----STATE 1--a----
	Exepected waiting time: 2.0
	s1 -> s0 : lambda = 0.3
	s1 -> s1 : lambda = 0.2
	"""
	if 'init' in labelling:
		msg =  "The label 'init' cannot be used: it is reserved for initial states."
		raise SyntaxError(msg)
	
	states = [i[0] for i in transitions]+[i[1] for i in transitions]
	states +=[i[0] for i in synchronous_transitions]+[i[2] for i in synchronous_transitions]
	states = list(set(states))
	states.sort()
	nb_states = len(states)
	
	if nb_states > len(labelling):
		raise ValueError("All states are not labelled (the labelling list is too small).")
	elif nb_states < len(labelling):
		print("WARNING: the labelling list is bigger than the number of states")

	res = zeros((nb_states,nb_states))
	for t in transitions:
		res[states.index(t[0])][states.index(t[1])] = t[2]
	
	labelling.append('init')
	res = vstack((res,zeros(len(res))))
	res = hstack((res,zeros(len(res))[:,newaxis]))
	if type(initial_state) == int:
		res[-1][initial_state] = 1.0
	else:
		if type(initial_state) == ndarray:
			initial_state = initial_state.tolist()
		res[-1] = array(initial_state+[0.0])
	return CTMC(res, labelling, name,synchronous_transitions)


def synchronousCompositionCTMCs(m1: CTMC, m2: CTMC, name: str = "unknown_composition") -> PCTMC:
	"""
	Returns the synchornous compotision of `m1` and `m2`.
	The output model is a PCTMC since some parameters are shared
	by multiple transitions.

	Parameters
	----------
	m1 : CTMC
		First CTMC to compose with.
	m2 : CTMC
		Second CTMC to compose with.

	Returns
	-------
	PCTMC
		Synchronous composition of `m1` and `m2`.
	"""
	m1_init = [i for i,li in enumerate(m1.labelling) if li == 'init']
	m2_init = [i for i,li in enumerate(m2.labelling) if li == 'init']
	m1_nb_states = m1.nb_states - len(m1_init)
	m2_nb_states = m2.nb_states - len(m2_init)
	nb_states = m1_nb_states * m2_nb_states	
	m1_sids = [i-m1.labelling[:i].count("init") for i,li in enumerate(m1.labelling) if li != 'init']
	m2_sids = [i-m1.labelling[:i].count("init") for i,li in enumerate(m2.labelling) if li != 'init']

	matrix= zeros((nb_states,nb_states),dtype='uint8')
	labelling = []
	p_v = {}
	p_str = []
	p_i = []
	t_e = [sympify('0.0')]

	def get_state_index(s1,s2):
		s2 = m2_sids.index(s2)
		s1 = m1_sids.index(s1)
		return s1*m2_nb_states+s2

	def add_in_matrix(trans,s1,s2,model,add_index=[]):
		if model == 1:
			for i in m2_sids:
				x,y = get_state_index(s1,i),get_state_index(s2,i)
				matrix[x,y] = trans
				for ind in add_index:
					p_i[ind].append([x,y])
		elif model == 2:
			for i in m1_sids:
				x,y = get_state_index(i,s1),get_state_index(i,s2)
				matrix[x,y] = trans
				for ind in add_index:
					p_i[ind].append([x,y])

	for s1 in m1_sids: # labelling
		for s2 in m2_sids:
			labelling.append(m1.labelling[s1]+','+m2.labelling[s2])

	for i in m1_sids: # m1 transitions
		for j in m1_sids:
			if m1.matrix[i,j] != 0.0:
				p_str.append("p_"+str(i)+'_'+str(j))
				p_v["p_"+str(i)+'_'+str(j)] = m1.matrix[i,j]
				p_i.append([])
				t_e.append(sympify(p_str[-1]))
				add_in_matrix(len(t_e)-1,s1=i,s2=j,model=1,add_index=[len(p_i)-1])
	for i in m2_sids: # m2 transitions
		for j in m2_sids:
			if m2.matrix[i,j] != 0.0:
				p_str.append("q_"+str(i)+'_'+str(j))
				p_v["q_"+str(i)+'_'+str(j)] = m2.matrix[i,j]
				p_i.append([])
				t_e.append(sympify(p_str[-1]))
				add_in_matrix(len(t_e)-1,s1=i,s2=j,model=2,add_index=[len(p_i)-1])

	for i in m1_sids: # self loops
		if m1.matrix[i,i] != 0.0:
			for j in m2_sids:
				if m2.matrix[j,j] != 0.0:
					matrix[get_state_index(i,j),get_state_index(i,j)] = len(t_e)
					t_e.append(sympify("p"+str(i)+'_'+str(i)+'+q'+str(j)+'_'+str(j)))

	for si,ai,di,pi in m1.synchronous_transitions: # synchronous transitions
		for sj,aj,dj,pj in m2.synchronous_transitions:
			if ai == aj:
				t_e.append(t_e[matrix[get_state_index(si,sj),get_state_index(di,dj)]]+sympify(pi*pj))
				matrix[get_state_index(si,sj),get_state_index(di,dj)] = len(t_e)-1
	
	labelling.append('init')
	matrix = vstack((matrix,zeros(nb_states,dtype='uint8')))
	matrix = hstack((matrix,zeros(nb_states+1,dtype='uint8')[:,newaxis]))
	m1_init_trans = zeros(m1_nb_states)
	for i in m1_init:
		tmp = [m1.matrix[i][j] for j in m1_sids]
		for j in range(m1_nb_states):
			if m1_init_trans[j]*tmp[j]>0.0:
				m1_init_trans[j] *= tmp[j]
			else:
				m1_init_trans[j] += tmp[j]
	m2_init_trans = zeros(m2_nb_states)
	for i in m2_init:
		tmp = [m2.matrix[i][j] for j in m2_sids]
		for j in range(m2_nb_states):
			if m2_init_trans[j]*tmp[j]>0.0:
				m2_init_trans[j] *= tmp[j]
			else:
				m2_init_trans[j] += tmp[j]
	for i,si in enumerate(m1_init_trans):
		if si > 0.0:
			p_str.append("initp_"+str(i))
			p_v["initp_"+str(i)] = si
			p_i.append([])
	for i,si in enumerate(m2_init_trans):
		if si > 0.0:
			p_str.append("initq_"+str(i))
			p_v["initq_"+str(i)] = si
			p_i.append([])
	for i,si in enumerate(m1_init_trans):
		for j,sj in enumerate(m2_init_trans):
			if si*sj <= 0.0:
				matrix[-1][get_state_index(i,j)] = 0
			else:
				matrix[-1][get_state_index(i,j)] = len(t_e)
				t_e.append(sympify('initp_'+str(i)+'*initq_'+str(j)))
				p_i[p_str.index('initp_'+str(i))].append([len(matrix)-1,get_state_index(i,j)])
				p_i[p_str.index('initq_'+str(j))].append([len(matrix)-1,get_state_index(i,j)])
	
	i = 0
	while i < len(matrix): # removing unreachable states
		if (matrix.T[i] == 0).all() == True and labelling[i] != 'init':
			nb_states -= 1
			matrix = delete(matrix,i,0)
			matrix = delete(matrix,i,1)
			labelling = labelling[:i]+labelling[i+1:]
			for j in range(len(p_i)):
				k = 0
				while k < len(p_i[j]):
					if p_i[j][k][0] == i or p_i[j][k][1] == i:
						p_i[j].remove(p_i[j][k])
					else:
						if p_i[j][k][0] > i:
							p_i[j][k][0] -= 1
						if p_i[j][k][1] > i:
							p_i[j][k][1] -= 1
						k += 1
			i = -1
		i += 1
	
	return PCTMC(matrix, labelling, t_e, p_v, p_i, p_str, name)




import unittest
from ..hmm import *
from os import remove
from ..base.Set import *
from math import log
from ..base.BW import BW

def modelHMM4():
	transitions = [(0,1,0.5),(0,2,0.5),(1,3,1.0),(2,4,1.0),
				   (3,0,0.8),(3,1,0.1),(3,2,0.1),(4,3,1.0)]
	emission = [(0,"x",0.4),(0,"y",0.6),(1,"a",0.8),(1,"b",0.2),
				(2,"a",0.1),(2,"b",0.9),(3,"x",0.5),(3,"y",0.5),
				(4,"y",1.0)]
	return createHMM(transitions,emission,0,"HMM4")

m = modelHMM4()

class HMMTestclass(unittest.TestCase):

	def test_HMM_state(var):
		var.assertEqual(m.a(0,0),0.0)
		var.assertEqual(m.a(0,1),0.5)
		var.assertEqual(m.b(0,'x'),0.4)
		var.assertEqual(m.b(0,'something else'),0.0)
		var.assertEqual(m.tau(0,1,'x'),0.2)
		var.assertEqual(set(m.getAlphabet(0)),set(['x','y']))
	
	def test_HMM_save_load_str(var):
		m.save("test_save.txt")
		mprime = loadHMM("test_save.txt")
		var.assertEqual(str(m),str(mprime))
		remove("test_save.txt")
	
	def test_HMM_observations(var):
		var.assertEqual(set(m.getAlphabet()),set(['x','y','a','b']))
	
	def test_HMM_Set(var):
		set1 = m.generateSet(50,10)
		set2 = m.generateSet(50,1/4,"geo",6)
		set1.addSet(set2)
		var.assertEqual(set1.type,0)
		set1.save("test_save.txt")
		set2 = loadSet("test_save.txt")
		var.assertTrue(set1.isEqual(set2))
		remove("test_save.txt")

	def test_HMM_logLikelihood(var):
		set1 = Set([['x','a','x','y']],[1])
		set2 = Set([['y','b','y','x']],[2])
		l11 = m._logLikelihood_oneproc(set1)
		l12 = m._logLikelihood_multiproc(set1)
		var.assertAlmostEqual(l11,l12)
		var.assertAlmostEqual(l11,log(0.0384))
		l2 = m.logLikelihood(set2)
		var.assertAlmostEqual(l2,log(0.1446))
		set1.addSet(set2)
		l3 = m.logLikelihood(set1)
		var.assertAlmostEqual(l3,(log(0.0384)+2*log(0.1446))/3)
	
	def test_BW_HMM(var):
		initial_model   = loadHMM("jajapy/tests/materials/hmm/random_HMM.txt")
		training_set    = loadSet("jajapy/tests/materials/hmm/training_set_HMM.txt")
		output_expected = loadHMM("jajapy/tests/materials/hmm/output_HMM.txt")
		output_gotten   = BW().fit( training_set, initial_model, stormpy_output=False)
		test_set = m.generateSet(10000,10)
		var.assertAlmostEqual(output_expected.logLikelihood(test_set),
							  output_gotten.logLikelihood(test_set))

if __name__ == "__main__":
	unittest.main()